/* Generated by Cython 0.25.2 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__fuzzysearch___generic_search
#define __PYX_HAVE_API__fuzzysearch___generic_search
#include <string.h>
#include <stdlib.h>
#include "memmem.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "src\\fuzzysearch\\_generic_search.pyx",
};

/*--- Type declarations ---*/
struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate;

/* "fuzzysearch/_generic_search.pyx":17
 * 
 * 
 * cdef struct GenericSearchCandidate:             # <<<<<<<<<<<<<<
 *     int start, subseq_index, l_dist, n_subs, n_ins, n_dels
 * 
 */
struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate {
  int start;
  int subseq_index;
  int l_dist;
  int n_subs;
  int n_ins;
  int n_dels;
};

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'fuzzysearch._generic_search' */
static PyObject *__pyx_f_11fuzzysearch_15_generic_search__c_find_near_matches_generic_linear_programming(char const *, size_t, char const *, size_t, unsigned int, unsigned int, unsigned int, unsigned int); /*proto*/
#define __Pyx_MODULE_NAME "fuzzysearch._generic_search"
int __pyx_module_is_main_fuzzysearch___generic_search = 0;

/* Implementation of 'fuzzysearch._generic_search' */
static PyObject *__pyx_builtin_ImportError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_xrange;
static const char __pyx_k_Seq[] = "Seq";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_end[] = "end";
static const char __pyx_k_six[] = "six";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_Match[] = "Match";
static const char __pyx_k_index[] = "index";
static const char __pyx_k_match[] = "match";
static const char __pyx_k_n_ins[] = "n_ins";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_l_dist[] = "l_dist";
static const char __pyx_k_n_dels[] = "n_dels";
static const char __pyx_k_n_subs[] = "n_subs";
static const char __pyx_k_xrange[] = "xrange";
static const char __pyx_k_Bio_Seq[] = "Bio.Seq";
static const char __pyx_k_matches[] = "matches";
static const char __pyx_k_replace[] = "_replace";
static const char __pyx_k_seq_len[] = "_seq_len";
static const char __pyx_k_sequence[] = "sequence";
static const char __pyx_k_unpacked[] = "unpacked";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_match_ptr[] = "match_ptr";
static const char __pyx_k_ngram_len[] = "ngram_len";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_c_sequence[] = "c_sequence";
static const char __pyx_k_max_l_dist[] = "max_l_dist";
static const char __pyx_k_subseq_len[] = "_subseq_len";
static const char __pyx_k_subseq_sum[] = "subseq_sum";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_binary_type[] = "binary_type";
static const char __pyx_k_ngram_start[] = "ngram_start";
static const char __pyx_k_subsequence[] = "subsequence";
static const char __pyx_k_c_max_l_dist[] = "c_max_l_dist";
static const char __pyx_k_subseq_index[] = "subseq_index";
static const char __pyx_k_ALLOWED_TYPES[] = "ALLOWED_TYPES";
static const char __pyx_k_c_subsequence[] = "c_subsequence";
static const char __pyx_k_max_deletions[] = "max_deletions";
static const char __pyx_k_search_params[] = "search_params";
static const char __pyx_k_max_insertions[] = "max_insertions";
static const char __pyx_k_c_max_deletions[] = "c_max_deletions";
static const char __pyx_k_c_max_insertions[] = "c_max_insertions";
static const char __pyx_k_max_substitutions[] = "max_substitutions";
static const char __pyx_k_fuzzysearch_common[] = "fuzzysearch.common";
static const char __pyx_k_c_max_substitutions[] = "c_max_substitutions";
static const char __pyx_k_small_search_length[] = "small_search_length";
static const char __pyx_k_subseq_len_minus_one[] = "_subseq_len_minus_one";
static const char __pyx_k_small_search_start_index[] = "small_search_start_index";
static const char __pyx_k_Given_subsequence_is_empty[] = "Given subsequence is empty!";
static const char __pyx_k_fuzzysearch__generic_search[] = "fuzzysearch._generic_search";
static const char __pyx_k_sequence_is_of_invalid_type_s[] = "sequence is of invalid type %s";
static const char __pyx_k_C_Users_taleinat_PycharmProjects[] = "C:\\Users\\taleinat\\PycharmProjects\\fuzzysearch\\src\\fuzzysearch\\_generic_search.pyx";
static const char __pyx_k_c_find_near_matches_generic_line[] = "c_find_near_matches_generic_linear_programming";
static const char __pyx_k_c_find_near_matches_generic_ngra[] = "c_find_near_matches_generic_ngrams";
static const char __pyx_k_subsequence_is_of_invalid_type_s[] = "subsequence is of invalid type %s";
static const char __pyx_k_the_subsequence_length_must_be_g[] = "the subsequence length must be greater than max_l_dist";
static PyObject *__pyx_n_s_ALLOWED_TYPES;
static PyObject *__pyx_n_s_Bio_Seq;
static PyObject *__pyx_kp_s_C_Users_taleinat_PycharmProjects;
static PyObject *__pyx_kp_s_Given_subsequence_is_empty;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_n_s_Match;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_n_s_Seq;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_binary_type;
static PyObject *__pyx_n_s_c_find_near_matches_generic_line;
static PyObject *__pyx_n_s_c_find_near_matches_generic_ngra;
static PyObject *__pyx_n_s_c_max_deletions;
static PyObject *__pyx_n_s_c_max_insertions;
static PyObject *__pyx_n_s_c_max_l_dist;
static PyObject *__pyx_n_s_c_max_substitutions;
static PyObject *__pyx_n_s_c_sequence;
static PyObject *__pyx_n_s_c_subsequence;
static PyObject *__pyx_n_s_end;
static PyObject *__pyx_n_s_fuzzysearch__generic_search;
static PyObject *__pyx_n_s_fuzzysearch_common;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_index;
static PyObject *__pyx_n_s_l_dist;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_match;
static PyObject *__pyx_n_s_match_ptr;
static PyObject *__pyx_n_s_matches;
static PyObject *__pyx_n_s_max_deletions;
static PyObject *__pyx_n_s_max_insertions;
static PyObject *__pyx_n_s_max_l_dist;
static PyObject *__pyx_n_s_max_substitutions;
static PyObject *__pyx_n_s_n_dels;
static PyObject *__pyx_n_s_n_ins;
static PyObject *__pyx_n_s_n_subs;
static PyObject *__pyx_n_s_ngram_len;
static PyObject *__pyx_n_s_ngram_start;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_replace;
static PyObject *__pyx_n_s_search_params;
static PyObject *__pyx_n_s_seq_len;
static PyObject *__pyx_n_s_sequence;
static PyObject *__pyx_kp_s_sequence_is_of_invalid_type_s;
static PyObject *__pyx_n_s_six;
static PyObject *__pyx_n_s_small_search_length;
static PyObject *__pyx_n_s_small_search_start_index;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_subseq_index;
static PyObject *__pyx_n_s_subseq_len;
static PyObject *__pyx_n_s_subseq_len_minus_one;
static PyObject *__pyx_n_s_subseq_sum;
static PyObject *__pyx_n_s_subsequence;
static PyObject *__pyx_kp_s_subsequence_is_of_invalid_type_s;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_kp_s_the_subsequence_length_must_be_g;
static PyObject *__pyx_n_s_unpacked;
static PyObject *__pyx_n_s_xrange;
static PyObject *__pyx_pf_11fuzzysearch_15_generic_search_c_find_near_matches_generic_linear_programming(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subsequence, PyObject *__pyx_v_sequence, PyObject *__pyx_v_search_params); /* proto */
static PyObject *__pyx_pf_11fuzzysearch_15_generic_search_2c_find_near_matches_generic_ngrams(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subsequence, PyObject *__pyx_v_sequence, PyObject *__pyx_v_search_params); /* proto */
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_536870912;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__2;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_codeobj__5;
static PyObject *__pyx_codeobj__7;

/* "fuzzysearch/_generic_search.pyx":30
 * 
 * 
 * def c_find_near_matches_generic_linear_programming(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11fuzzysearch_15_generic_search_1c_find_near_matches_generic_linear_programming(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11fuzzysearch_15_generic_search_c_find_near_matches_generic_linear_programming[] = "search for near-matches of subsequence in sequence\n\n    This searches for near-matches, where the nearly-matching parts of the\n    sequence must meet the following limitations (relative to the subsequence):\n\n    * the maximum allowed number of character substitutions\n    * the maximum allowed number of new characters inserted\n    * and the maximum allowed number of character deletions\n    * the total number of substitutions, insertions and deletions\n    ";
static PyMethodDef __pyx_mdef_11fuzzysearch_15_generic_search_1c_find_near_matches_generic_linear_programming = {"c_find_near_matches_generic_linear_programming", (PyCFunction)__pyx_pw_11fuzzysearch_15_generic_search_1c_find_near_matches_generic_linear_programming, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11fuzzysearch_15_generic_search_c_find_near_matches_generic_linear_programming};
static PyObject *__pyx_pw_11fuzzysearch_15_generic_search_1c_find_near_matches_generic_linear_programming(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_subsequence = 0;
  PyObject *__pyx_v_sequence = 0;
  PyObject *__pyx_v_search_params = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_find_near_matches_generic_linear_programming (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_subsequence,&__pyx_n_s_sequence,&__pyx_n_s_search_params,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_subsequence)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_sequence)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_linear_programming", 1, 3, 3, 1); __PYX_ERR(0, 30, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_search_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_linear_programming", 1, 3, 3, 2); __PYX_ERR(0, 30, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "c_find_near_matches_generic_linear_programming") < 0)) __PYX_ERR(0, 30, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_subsequence = values[0];
    __pyx_v_sequence = values[1];
    __pyx_v_search_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_linear_programming", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 30, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fuzzysearch._generic_search.c_find_near_matches_generic_linear_programming", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11fuzzysearch_15_generic_search_c_find_near_matches_generic_linear_programming(__pyx_self, __pyx_v_subsequence, __pyx_v_sequence, __pyx_v_search_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11fuzzysearch_15_generic_search_c_find_near_matches_generic_linear_programming(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subsequence, PyObject *__pyx_v_sequence, PyObject *__pyx_v_search_params) {
  PyObject *__pyx_v_max_substitutions = NULL;
  PyObject *__pyx_v_max_insertions = NULL;
  PyObject *__pyx_v_max_deletions = NULL;
  PyObject *__pyx_v_max_l_dist = NULL;
  char const *__pyx_v_c_subsequence;
  char const *__pyx_v_c_sequence;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  char const *__pyx_t_10;
  char const *__pyx_t_11;
  Py_ssize_t __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  unsigned int __pyx_t_14;
  unsigned int __pyx_t_15;
  unsigned int __pyx_t_16;
  unsigned int __pyx_t_17;
  unsigned int __pyx_t_18;
  __Pyx_RefNannySetupContext("c_find_near_matches_generic_linear_programming", 0);

  /* "fuzzysearch/_generic_search.pyx":41
 *     * the total number of substitutions, insertions and deletions
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ALLOWED_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_sequence, __pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "fuzzysearch/_generic_search.pyx":42
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))             # <<<<<<<<<<<<<<
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 */
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_sequence_is_of_invalid_type_s, ((PyObject *)Py_TYPE(__pyx_v_subsequence))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 42, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 42, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":41
 *     * the total number of substitutions, insertions and deletions
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 */
  }

  /* "fuzzysearch/_generic_search.pyx":43
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ALLOWED_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_subsequence, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_2) {

    /* "fuzzysearch/_generic_search.pyx":44
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))             # <<<<<<<<<<<<<<
 * 
 *     if not subsequence:
 */
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_subsequence_is_of_invalid_type_s, ((PyObject *)Py_TYPE(__pyx_v_subsequence))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 44, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 44, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":43
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 */
  }

  /* "fuzzysearch/_generic_search.pyx":46
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 *     if not subsequence:             # <<<<<<<<<<<<<<
 *         raise ValueError('Given subsequence is empty!')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_subsequence); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 46, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "fuzzysearch/_generic_search.pyx":47
 * 
 *     if not subsequence:
 *         raise ValueError('Given subsequence is empty!')             # <<<<<<<<<<<<<<
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple_, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 47, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 47, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":46
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 *     if not subsequence:             # <<<<<<<<<<<<<<
 *         raise ValueError('Given subsequence is empty!')
 * 
 */
  }

  /* "fuzzysearch/_generic_search.pyx":49
 *         raise ValueError('Given subsequence is empty!')
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked             # <<<<<<<<<<<<<<
 * 
 *     cdef const char *c_subsequence = subsequence
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_search_params, __pyx_n_s_unpacked); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 49, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 49, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7};
    __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 49, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_9(__pyx_t_8); if (unlikely(!item)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 4) < 0) __PYX_ERR(0, 49, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 49, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_max_substitutions = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_max_insertions = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_max_deletions = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_max_l_dist = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "fuzzysearch/_generic_search.pyx":51
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked
 * 
 *     cdef const char *c_subsequence = subsequence             # <<<<<<<<<<<<<<
 *     cdef const char *c_sequence = sequence
 * 
 */
  __pyx_t_10 = __Pyx_PyObject_AsString(__pyx_v_subsequence); if (unlikely((!__pyx_t_10) && PyErr_Occurred())) __PYX_ERR(0, 51, __pyx_L1_error)
  __pyx_v_c_subsequence = __pyx_t_10;

  /* "fuzzysearch/_generic_search.pyx":52
 * 
 *     cdef const char *c_subsequence = subsequence
 *     cdef const char *c_sequence = sequence             # <<<<<<<<<<<<<<
 * 
 *     return _c_find_near_matches_generic_linear_programming(
 */
  __pyx_t_11 = __Pyx_PyObject_AsString(__pyx_v_sequence); if (unlikely((!__pyx_t_11) && PyErr_Occurred())) __PYX_ERR(0, 52, __pyx_L1_error)
  __pyx_v_c_sequence = __pyx_t_11;

  /* "fuzzysearch/_generic_search.pyx":54
 *     cdef const char *c_sequence = sequence
 * 
 *     return _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *         c_subsequence, len(subsequence),
 *         c_sequence, len(sequence),
 */
  __Pyx_XDECREF(__pyx_r);

  /* "fuzzysearch/_generic_search.pyx":55
 * 
 *     return _c_find_near_matches_generic_linear_programming(
 *         c_subsequence, len(subsequence),             # <<<<<<<<<<<<<<
 *         c_sequence, len(sequence),
 *         max_substitutions if max_substitutions is not None else (1<<29),
 */
  __pyx_t_12 = PyObject_Length(__pyx_v_subsequence); if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 55, __pyx_L1_error)

  /* "fuzzysearch/_generic_search.pyx":56
 *     return _c_find_near_matches_generic_linear_programming(
 *         c_subsequence, len(subsequence),
 *         c_sequence, len(sequence),             # <<<<<<<<<<<<<<
 *         max_substitutions if max_substitutions is not None else (1<<29),
 *         max_insertions if max_insertions is not None else (1<<29),
 */
  __pyx_t_13 = PyObject_Length(__pyx_v_sequence); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 56, __pyx_L1_error)

  /* "fuzzysearch/_generic_search.pyx":57
 *         c_subsequence, len(subsequence),
 *         c_sequence, len(sequence),
 *         max_substitutions if max_substitutions is not None else (1<<29),             # <<<<<<<<<<<<<<
 *         max_insertions if max_insertions is not None else (1<<29),
 *         max_deletions if max_deletions is not None else (1<<29),
 */
  __pyx_t_3 = (__pyx_v_max_substitutions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_15 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_substitutions); if (unlikely((__pyx_t_15 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 57, __pyx_L1_error)
    __pyx_t_14 = __pyx_t_15;
  } else {
    __pyx_t_14 = 0x20000000;
  }

  /* "fuzzysearch/_generic_search.pyx":58
 *         c_sequence, len(sequence),
 *         max_substitutions if max_substitutions is not None else (1<<29),
 *         max_insertions if max_insertions is not None else (1<<29),             # <<<<<<<<<<<<<<
 *         max_deletions if max_deletions is not None else (1<<29),
 *         max_l_dist if max_l_dist is not None else (1<<29),
 */
  __pyx_t_3 = (__pyx_v_max_insertions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_16 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_insertions); if (unlikely((__pyx_t_16 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 58, __pyx_L1_error)
    __pyx_t_15 = __pyx_t_16;
  } else {
    __pyx_t_15 = 0x20000000;
  }

  /* "fuzzysearch/_generic_search.pyx":59
 *         max_substitutions if max_substitutions is not None else (1<<29),
 *         max_insertions if max_insertions is not None else (1<<29),
 *         max_deletions if max_deletions is not None else (1<<29),             # <<<<<<<<<<<<<<
 *         max_l_dist if max_l_dist is not None else (1<<29),
 *     )
 */
  __pyx_t_3 = (__pyx_v_max_deletions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_17 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_deletions); if (unlikely((__pyx_t_17 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 59, __pyx_L1_error)
    __pyx_t_16 = __pyx_t_17;
  } else {
    __pyx_t_16 = 0x20000000;
  }

  /* "fuzzysearch/_generic_search.pyx":60
 *         max_insertions if max_insertions is not None else (1<<29),
 *         max_deletions if max_deletions is not None else (1<<29),
 *         max_l_dist if max_l_dist is not None else (1<<29),             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_3 = (__pyx_v_max_l_dist != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_18 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_l_dist); if (unlikely((__pyx_t_18 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 60, __pyx_L1_error)
    __pyx_t_17 = __pyx_t_18;
  } else {
    __pyx_t_17 = 0x20000000;
  }

  /* "fuzzysearch/_generic_search.pyx":54
 *     cdef const char *c_sequence = sequence
 * 
 *     return _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *         c_subsequence, len(subsequence),
 *         c_sequence, len(sequence),
 */
  __pyx_t_1 = __pyx_f_11fuzzysearch_15_generic_search__c_find_near_matches_generic_linear_programming(__pyx_v_c_subsequence, __pyx_t_12, __pyx_v_c_sequence, __pyx_t_13, __pyx_t_14, __pyx_t_15, __pyx_t_16, __pyx_t_17); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "fuzzysearch/_generic_search.pyx":30
 * 
 * 
 * def c_find_near_matches_generic_linear_programming(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("fuzzysearch._generic_search.c_find_near_matches_generic_linear_programming", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_max_substitutions);
  __Pyx_XDECREF(__pyx_v_max_insertions);
  __Pyx_XDECREF(__pyx_v_max_deletions);
  __Pyx_XDECREF(__pyx_v_max_l_dist);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fuzzysearch/_generic_search.pyx":66
 * # subsequence strings, which means if they contain null bytes the data after
 * # the first null byte will not be copied.
 * cdef _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *         const char* subsequence, size_t subseq_len,
 *         const char* sequence, size_t seq_len,
 */

static PyObject *__pyx_f_11fuzzysearch_15_generic_search__c_find_near_matches_generic_linear_programming(char const *__pyx_v_subsequence, size_t __pyx_v_subseq_len, char const *__pyx_v_sequence, size_t __pyx_v_seq_len, unsigned int __pyx_v_max_substitutions, unsigned int __pyx_v_max_insertions, unsigned int __pyx_v_max_deletions, unsigned int __pyx_v_max_l_dist) {
  unsigned int __pyx_v_subseq_len_minus_one;
  size_t __pyx_v_alloc_size;
  struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *__pyx_v_candidates;
  struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *__pyx_v_new_candidates;
  struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *__pyx_v__tmp;
  struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate __pyx_v_cand;
  size_t __pyx_v_n_candidates;
  size_t __pyx_v_n_new_candidates;
  size_t __pyx_v_n_cand;
  PyObject *__pyx_v_matches = NULL;
  size_t __pyx_v_index;
  char __pyx_v_seq_char;
  int __pyx_v_have_realloced;
  PyObject *__pyx_v_n_skipped = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  long __pyx_t_2;
  size_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  char *__pyx_t_7;
  char *__pyx_t_8;
  char *__pyx_t_9;
  struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_t_17;
  PyObject *__pyx_t_18 = NULL;
  int __pyx_t_19;
  int __pyx_t_20;
  unsigned int __pyx_t_21;
  unsigned int __pyx_t_22;
  unsigned int __pyx_t_23;
  Py_ssize_t __pyx_t_24;
  PyObject *(*__pyx_t_25)(PyObject *);
  PyObject *__pyx_t_26 = NULL;
  Py_ssize_t __pyx_t_27;
  int __pyx_t_28;
  char const *__pyx_t_29;
  PyObject *__pyx_t_30 = NULL;
  PyObject *__pyx_t_31 = NULL;
  PyObject *__pyx_t_32 = NULL;
  PyObject *__pyx_t_33 = NULL;
  PyObject *__pyx_t_34 = NULL;
  PyObject *__pyx_t_35 = NULL;
  __Pyx_RefNannySetupContext("_c_find_near_matches_generic_linear_programming", 0);

  /* "fuzzysearch/_generic_search.pyx":74
 *         unsigned int max_l_dist,
 * ):
 *     cdef unsigned int subseq_len_minus_one = subseq_len - 1             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t alloc_size
 */
  __pyx_v_subseq_len_minus_one = (__pyx_v_subseq_len - 1);

  /* "fuzzysearch/_generic_search.pyx":81
 *     cdef GenericSearchCandidate* _tmp
 *     cdef GenericSearchCandidate cand
 *     cdef size_t n_candidates = 0             # <<<<<<<<<<<<<<
 *     cdef size_t n_new_candidates = 0
 *     cdef size_t n_cand
 */
  __pyx_v_n_candidates = 0;

  /* "fuzzysearch/_generic_search.pyx":82
 *     cdef GenericSearchCandidate cand
 *     cdef size_t n_candidates = 0
 *     cdef size_t n_new_candidates = 0             # <<<<<<<<<<<<<<
 *     cdef size_t n_cand
 * 
 */
  __pyx_v_n_new_candidates = 0;

  /* "fuzzysearch/_generic_search.pyx":85
 *     cdef size_t n_cand
 * 
 *     alloc_size = min(10, subseq_len * 3 + 1)             # <<<<<<<<<<<<<<
 *     candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:
 */
  __pyx_t_1 = ((__pyx_v_subseq_len * 3) + 1);
  __pyx_t_2 = 10;
  if (((__pyx_t_1 < __pyx_t_2) != 0)) {
    __pyx_t_3 = __pyx_t_1;
  } else {
    __pyx_t_3 = __pyx_t_2;
  }
  __pyx_v_alloc_size = __pyx_t_3;

  /* "fuzzysearch/_generic_search.pyx":86
 * 
 *     alloc_size = min(10, subseq_len * 3 + 1)
 *     candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))             # <<<<<<<<<<<<<<
 *     if candidates is NULL:
 *         raise MemoryError()
 */
  __pyx_v_candidates = ((struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *)malloc((__pyx_v_alloc_size * (sizeof(struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate)))));

  /* "fuzzysearch/_generic_search.pyx":87
 *     alloc_size = min(10, subseq_len * 3 + 1)
 *     candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 */
  __pyx_t_4 = ((__pyx_v_candidates == NULL) != 0);
  if (__pyx_t_4) {

    /* "fuzzysearch/_generic_search.pyx":88
 *     candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:
 */
    PyErr_NoMemory(); __PYX_ERR(0, 88, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":87
 *     alloc_size = min(10, subseq_len * 3 + 1)
 *     candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 */
  }

  /* "fuzzysearch/_generic_search.pyx":89
 *     if candidates is NULL:
 *         raise MemoryError()
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))             # <<<<<<<<<<<<<<
 *     if candidates is NULL:
 *         free(candidates)
 */
  __pyx_v_new_candidates = ((struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *)malloc((__pyx_v_alloc_size * (sizeof(struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate)))));

  /* "fuzzysearch/_generic_search.pyx":90
 *         raise MemoryError()
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:             # <<<<<<<<<<<<<<
 *         free(candidates)
 *         raise MemoryError()
 */
  __pyx_t_4 = ((__pyx_v_candidates == NULL) != 0);
  if (__pyx_t_4) {

    /* "fuzzysearch/_generic_search.pyx":91
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:
 *         free(candidates)             # <<<<<<<<<<<<<<
 *         raise MemoryError()
 * 
 */
    free(__pyx_v_candidates);

    /* "fuzzysearch/_generic_search.pyx":92
 *     if candidates is NULL:
 *         free(candidates)
 *         raise MemoryError()             # <<<<<<<<<<<<<<
 * 
 *     matches = []
 */
    PyErr_NoMemory(); __PYX_ERR(0, 92, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":90
 *         raise MemoryError()
 *     new_candidates = <GenericSearchCandidate *> malloc(alloc_size * sizeof(GenericSearchCandidate))
 *     if candidates is NULL:             # <<<<<<<<<<<<<<
 *         free(candidates)
 *         raise MemoryError()
 */
  }

  /* "fuzzysearch/_generic_search.pyx":94
 *         raise MemoryError()
 * 
 *     matches = []             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t index
 */
  __pyx_t_5 = PyList_New(0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_v_matches = ((PyObject*)__pyx_t_5);
  __pyx_t_5 = 0;

  /* "fuzzysearch/_generic_search.pyx":99
 *     cdef char seq_char
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         index = 0
 *         have_realloced = False
 */
  /*try:*/ {

    /* "fuzzysearch/_generic_search.pyx":100
 * 
 *     try:
 *         index = 0             # <<<<<<<<<<<<<<
 *         have_realloced = False
 *         for seq_char in sequence[:seq_len]:
 */
    __pyx_v_index = 0;

    /* "fuzzysearch/_generic_search.pyx":101
 *     try:
 *         index = 0
 *         have_realloced = False             # <<<<<<<<<<<<<<
 *         for seq_char in sequence[:seq_len]:
 *             candidates[n_candidates] = GenericSearchCandidate(index, 0, 0, 0, 0, 0)
 */
    __pyx_v_have_realloced = 0;

    /* "fuzzysearch/_generic_search.pyx":102
 *         index = 0
 *         have_realloced = False
 *         for seq_char in sequence[:seq_len]:             # <<<<<<<<<<<<<<
 *             candidates[n_candidates] = GenericSearchCandidate(index, 0, 0, 0, 0, 0)
 *             n_candidates += 1
 */
    __pyx_t_5 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_sequence + 0, __pyx_v_seq_len - 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 102, __pyx_L6_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = PyBytes_AS_STRING(__pyx_t_5);
    __pyx_t_8 = (__pyx_t_7 + PyBytes_GET_SIZE(__pyx_t_5));
    for (__pyx_t_9 = __pyx_t_7; __pyx_t_9 < __pyx_t_8; __pyx_t_9++) {
      __pyx_t_6 = __pyx_t_9;
      __pyx_v_seq_char = (__pyx_t_6[0]);

      /* "fuzzysearch/_generic_search.pyx":103
 *         have_realloced = False
 *         for seq_char in sequence[:seq_len]:
 *             candidates[n_candidates] = GenericSearchCandidate(index, 0, 0, 0, 0, 0)             # <<<<<<<<<<<<<<
 *             n_candidates += 1
 * 
 */
      __pyx_t_10.start = __pyx_v_index;
      __pyx_t_10.subseq_index = 0;
      __pyx_t_10.l_dist = 0;
      __pyx_t_10.n_subs = 0;
      __pyx_t_10.n_ins = 0;
      __pyx_t_10.n_dels = 0;
      (__pyx_v_candidates[__pyx_v_n_candidates]) = __pyx_t_10;

      /* "fuzzysearch/_generic_search.pyx":104
 *         for seq_char in sequence[:seq_len]:
 *             candidates[n_candidates] = GenericSearchCandidate(index, 0, 0, 0, 0, 0)
 *             n_candidates += 1             # <<<<<<<<<<<<<<
 * 
 *             for n_cand in xrange(n_candidates):
 */
      __pyx_v_n_candidates = (__pyx_v_n_candidates + 1);

      /* "fuzzysearch/_generic_search.pyx":106
 *             n_candidates += 1
 * 
 *             for n_cand in xrange(n_candidates):             # <<<<<<<<<<<<<<
 *                 cand = candidates[n_cand]
 * 
 */
      __pyx_t_3 = __pyx_v_n_candidates;
      for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_3; __pyx_t_1+=1) {
        __pyx_v_n_cand = __pyx_t_1;

        /* "fuzzysearch/_generic_search.pyx":107
 * 
 *             for n_cand in xrange(n_candidates):
 *                 cand = candidates[n_cand]             # <<<<<<<<<<<<<<
 * 
 *                 if n_new_candidates + 4 > alloc_size:
 */
        __pyx_v_cand = (__pyx_v_candidates[__pyx_v_n_cand]);

        /* "fuzzysearch/_generic_search.pyx":109
 *                 cand = candidates[n_cand]
 * 
 *                 if n_new_candidates + 4 > alloc_size:             # <<<<<<<<<<<<<<
 *                     alloc_size *= 2
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 */
        __pyx_t_4 = (((__pyx_v_n_new_candidates + 4) > __pyx_v_alloc_size) != 0);
        if (__pyx_t_4) {

          /* "fuzzysearch/_generic_search.pyx":110
 * 
 *                 if n_new_candidates + 4 > alloc_size:
 *                     alloc_size *= 2             # <<<<<<<<<<<<<<
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                     if _tmp is NULL:
 */
          __pyx_v_alloc_size = (__pyx_v_alloc_size * 2);

          /* "fuzzysearch/_generic_search.pyx":111
 *                 if n_new_candidates + 4 > alloc_size:
 *                     alloc_size *= 2
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))             # <<<<<<<<<<<<<<
 *                     if _tmp is NULL:
 *                         raise MemoryError()
 */
          __pyx_v__tmp = ((struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *)realloc(__pyx_v_new_candidates, (__pyx_v_alloc_size * (sizeof(struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate)))));

          /* "fuzzysearch/_generic_search.pyx":112
 *                     alloc_size *= 2
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                     if _tmp is NULL:             # <<<<<<<<<<<<<<
 *                         raise MemoryError()
 *                     new_candidates = _tmp
 */
          __pyx_t_4 = ((__pyx_v__tmp == NULL) != 0);
          if (__pyx_t_4) {

            /* "fuzzysearch/_generic_search.pyx":113
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                     if _tmp is NULL:
 *                         raise MemoryError()             # <<<<<<<<<<<<<<
 *                     new_candidates = _tmp
 *                     have_realloced = True
 */
            PyErr_NoMemory(); __PYX_ERR(0, 113, __pyx_L6_error)

            /* "fuzzysearch/_generic_search.pyx":112
 *                     alloc_size *= 2
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                     if _tmp is NULL:             # <<<<<<<<<<<<<<
 *                         raise MemoryError()
 *                     new_candidates = _tmp
 */
          }

          /* "fuzzysearch/_generic_search.pyx":114
 *                     if _tmp is NULL:
 *                         raise MemoryError()
 *                     new_candidates = _tmp             # <<<<<<<<<<<<<<
 *                     have_realloced = True
 * 
 */
          __pyx_v_new_candidates = __pyx_v__tmp;

          /* "fuzzysearch/_generic_search.pyx":115
 *                         raise MemoryError()
 *                     new_candidates = _tmp
 *                     have_realloced = True             # <<<<<<<<<<<<<<
 * 
 *                 # if this sequence char is the candidate's next expected char
 */
          __pyx_v_have_realloced = 1;

          /* "fuzzysearch/_generic_search.pyx":109
 *                 cand = candidates[n_cand]
 * 
 *                 if n_new_candidates + 4 > alloc_size:             # <<<<<<<<<<<<<<
 *                     alloc_size *= 2
 *                     _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 */
        }

        /* "fuzzysearch/_generic_search.pyx":118
 * 
 *                 # if this sequence char is the candidate's next expected char
 *                 if seq_char == subsequence[cand.subseq_index]:             # <<<<<<<<<<<<<<
 *                     # if reached the end of the subsequence, return a match
 *                     if cand.subseq_index == subseq_len_minus_one:
 */
        __pyx_t_4 = ((__pyx_v_seq_char == (__pyx_v_subsequence[__pyx_v_cand.subseq_index])) != 0);
        if (__pyx_t_4) {

          /* "fuzzysearch/_generic_search.pyx":120
 *                 if seq_char == subsequence[cand.subseq_index]:
 *                     # if reached the end of the subsequence, return a match
 *                     if cand.subseq_index == subseq_len_minus_one:             # <<<<<<<<<<<<<<
 *                         matches.append(Match(cand.start, index + 1, cand.l_dist))
 *                     # otherwise, update the candidate's subseq_index and keep it
 */
          __pyx_t_4 = ((__pyx_v_cand.subseq_index == __pyx_v_subseq_len_minus_one) != 0);
          if (__pyx_t_4) {

            /* "fuzzysearch/_generic_search.pyx":121
 *                     # if reached the end of the subsequence, return a match
 *                     if cand.subseq_index == subseq_len_minus_one:
 *                         matches.append(Match(cand.start, index + 1, cand.l_dist))             # <<<<<<<<<<<<<<
 *                     # otherwise, update the candidate's subseq_index and keep it
 *                     else:
 */
            __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_Match); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 121, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_13 = __Pyx_PyInt_From_int(__pyx_v_cand.start); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 121, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_13);
            __pyx_t_14 = __Pyx_PyInt_FromSize_t((__pyx_v_index + 1)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 121, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 121, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_15);
            __pyx_t_16 = NULL;
            __pyx_t_17 = 0;
            if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
              __pyx_t_16 = PyMethod_GET_SELF(__pyx_t_12);
              if (likely(__pyx_t_16)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                __Pyx_INCREF(__pyx_t_16);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_12, function);
                __pyx_t_17 = 1;
              }
            }
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(__pyx_t_12)) {
              PyObject *__pyx_temp[4] = {__pyx_t_16, __pyx_t_13, __pyx_t_14, __pyx_t_15};
              __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 121, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
              PyObject *__pyx_temp[4] = {__pyx_t_16, __pyx_t_13, __pyx_t_14, __pyx_t_15};
              __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 121, __pyx_L6_error)
              __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
            } else
            #endif
            {
              __pyx_t_18 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 121, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_18);
              if (__pyx_t_16) {
                __Pyx_GIVEREF(__pyx_t_16); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_16); __pyx_t_16 = NULL;
              }
              __Pyx_GIVEREF(__pyx_t_13);
              PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_17, __pyx_t_13);
              __Pyx_GIVEREF(__pyx_t_14);
              PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_17, __pyx_t_14);
              __Pyx_GIVEREF(__pyx_t_15);
              PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_17, __pyx_t_15);
              __pyx_t_13 = 0;
              __pyx_t_14 = 0;
              __pyx_t_15 = 0;
              __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_18, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 121, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
            }
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_11); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 121, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

            /* "fuzzysearch/_generic_search.pyx":120
 *                 if seq_char == subsequence[cand.subseq_index]:
 *                     # if reached the end of the subsequence, return a match
 *                     if cand.subseq_index == subseq_len_minus_one:             # <<<<<<<<<<<<<<
 *                         matches.append(Match(cand.start, index + 1, cand.l_dist))
 *                     # otherwise, update the candidate's subseq_index and keep it
 */
            goto __pyx_L15;
          }

          /* "fuzzysearch/_generic_search.pyx":124
 *                     # otherwise, update the candidate's subseq_index and keep it
 *                     else:
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                             cand.start, cand.subseq_index + 1,
 *                             cand.l_dist, cand.n_subs,
 */
          /*else*/ {

            /* "fuzzysearch/_generic_search.pyx":125
 *                     else:
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                             cand.start, cand.subseq_index + 1,             # <<<<<<<<<<<<<<
 *                             cand.l_dist, cand.n_subs,
 *                             cand.n_ins, cand.n_dels,
 */
            __pyx_t_10.start = __pyx_v_cand.start;
            __pyx_t_10.subseq_index = (__pyx_v_cand.subseq_index + 1);

            /* "fuzzysearch/_generic_search.pyx":126
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                             cand.start, cand.subseq_index + 1,
 *                             cand.l_dist, cand.n_subs,             # <<<<<<<<<<<<<<
 *                             cand.n_ins, cand.n_dels,
 *                         )
 */
            __pyx_t_10.l_dist = __pyx_v_cand.l_dist;
            __pyx_t_10.n_subs = __pyx_v_cand.n_subs;

            /* "fuzzysearch/_generic_search.pyx":127
 *                             cand.start, cand.subseq_index + 1,
 *                             cand.l_dist, cand.n_subs,
 *                             cand.n_ins, cand.n_dels,             # <<<<<<<<<<<<<<
 *                         )
 *                         n_new_candidates += 1
 */
            __pyx_t_10.n_ins = __pyx_v_cand.n_ins;
            __pyx_t_10.n_dels = __pyx_v_cand.n_dels;

            /* "fuzzysearch/_generic_search.pyx":124
 *                     # otherwise, update the candidate's subseq_index and keep it
 *                     else:
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                             cand.start, cand.subseq_index + 1,
 *                             cand.l_dist, cand.n_subs,
 */
            (__pyx_v_new_candidates[__pyx_v_n_new_candidates]) = __pyx_t_10;

            /* "fuzzysearch/_generic_search.pyx":129
 *                             cand.n_ins, cand.n_dels,
 *                         )
 *                         n_new_candidates += 1             # <<<<<<<<<<<<<<
 * 
 *                 # if this sequence char is *not* the candidate's next expected char
 */
            __pyx_v_n_new_candidates = (__pyx_v_n_new_candidates + 1);
          }
          __pyx_L15:;

          /* "fuzzysearch/_generic_search.pyx":118
 * 
 *                 # if this sequence char is the candidate's next expected char
 *                 if seq_char == subsequence[cand.subseq_index]:             # <<<<<<<<<<<<<<
 *                     # if reached the end of the subsequence, return a match
 *                     if cand.subseq_index == subseq_len_minus_one:
 */
          goto __pyx_L14;
        }

        /* "fuzzysearch/_generic_search.pyx":136
 *                     # unless this candidate has already skipped the maximum allowed
 *                     # number of characters
 *                     if cand.l_dist == max_l_dist:             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
        /*else*/ {
          __pyx_t_4 = ((__pyx_v_cand.l_dist == __pyx_v_max_l_dist) != 0);
          if (__pyx_t_4) {

            /* "fuzzysearch/_generic_search.pyx":137
 *                     # number of characters
 *                     if cand.l_dist == max_l_dist:
 *                         continue             # <<<<<<<<<<<<<<
 * 
 *                     if cand.n_ins < max_insertions:
 */
            goto __pyx_L10_continue;

            /* "fuzzysearch/_generic_search.pyx":136
 *                     # unless this candidate has already skipped the maximum allowed
 *                     # number of characters
 *                     if cand.l_dist == max_l_dist:             # <<<<<<<<<<<<<<
 *                         continue
 * 
 */
          }

          /* "fuzzysearch/_generic_search.pyx":139
 *                         continue
 * 
 *                     if cand.n_ins < max_insertions:             # <<<<<<<<<<<<<<
 *                         # add a candidate skipping a sequence char
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 */
          __pyx_t_4 = ((__pyx_v_cand.n_ins < __pyx_v_max_insertions) != 0);
          if (__pyx_t_4) {

            /* "fuzzysearch/_generic_search.pyx":142
 *                         # add a candidate skipping a sequence char
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                             cand.start, cand.subseq_index,             # <<<<<<<<<<<<<<
 *                             cand.l_dist + 1, cand.n_subs,
 *                             cand.n_ins + 1, cand.n_dels,
 */
            __pyx_t_10.start = __pyx_v_cand.start;
            __pyx_t_10.subseq_index = __pyx_v_cand.subseq_index;

            /* "fuzzysearch/_generic_search.pyx":143
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                             cand.start, cand.subseq_index,
 *                             cand.l_dist + 1, cand.n_subs,             # <<<<<<<<<<<<<<
 *                             cand.n_ins + 1, cand.n_dels,
 *                         )
 */
            __pyx_t_10.l_dist = (__pyx_v_cand.l_dist + 1);
            __pyx_t_10.n_subs = __pyx_v_cand.n_subs;

            /* "fuzzysearch/_generic_search.pyx":144
 *                             cand.start, cand.subseq_index,
 *                             cand.l_dist + 1, cand.n_subs,
 *                             cand.n_ins + 1, cand.n_dels,             # <<<<<<<<<<<<<<
 *                         )
 *                         n_new_candidates += 1
 */
            __pyx_t_10.n_ins = (__pyx_v_cand.n_ins + 1);
            __pyx_t_10.n_dels = __pyx_v_cand.n_dels;

            /* "fuzzysearch/_generic_search.pyx":141
 *                     if cand.n_ins < max_insertions:
 *                         # add a candidate skipping a sequence char
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                             cand.start, cand.subseq_index,
 *                             cand.l_dist + 1, cand.n_subs,
 */
            (__pyx_v_new_candidates[__pyx_v_n_new_candidates]) = __pyx_t_10;

            /* "fuzzysearch/_generic_search.pyx":146
 *                             cand.n_ins + 1, cand.n_dels,
 *                         )
 *                         n_new_candidates += 1             # <<<<<<<<<<<<<<
 * 
 *                     if cand.subseq_index + 1 < subseq_len:
 */
            __pyx_v_n_new_candidates = (__pyx_v_n_new_candidates + 1);

            /* "fuzzysearch/_generic_search.pyx":139
 *                         continue
 * 
 *                     if cand.n_ins < max_insertions:             # <<<<<<<<<<<<<<
 *                         # add a candidate skipping a sequence char
 *                         new_candidates[n_new_candidates] = GenericSearchCandidate(
 */
          }

          /* "fuzzysearch/_generic_search.pyx":148
 *                         n_new_candidates += 1
 * 
 *                     if cand.subseq_index + 1 < subseq_len:             # <<<<<<<<<<<<<<
 *                         if cand.n_subs < max_substitutions:
 *                             # add a candidate skipping both a sequence char and a
 */
          __pyx_t_4 = (((__pyx_v_cand.subseq_index + 1) < __pyx_v_subseq_len) != 0);
          if (__pyx_t_4) {

            /* "fuzzysearch/_generic_search.pyx":149
 * 
 *                     if cand.subseq_index + 1 < subseq_len:
 *                         if cand.n_subs < max_substitutions:             # <<<<<<<<<<<<<<
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 */
            __pyx_t_4 = ((__pyx_v_cand.n_subs < __pyx_v_max_substitutions) != 0);
            if (__pyx_t_4) {

              /* "fuzzysearch/_generic_search.pyx":153
 *                             # subsequence char
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                 cand.start, cand.subseq_index + 1,             # <<<<<<<<<<<<<<
 *                                 cand.l_dist + 1, cand.n_subs + 1,
 *                                 cand.n_ins, cand.n_dels,
 */
              __pyx_t_10.start = __pyx_v_cand.start;
              __pyx_t_10.subseq_index = (__pyx_v_cand.subseq_index + 1);

              /* "fuzzysearch/_generic_search.pyx":154
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs + 1,             # <<<<<<<<<<<<<<
 *                                 cand.n_ins, cand.n_dels,
 *                             )
 */
              __pyx_t_10.l_dist = (__pyx_v_cand.l_dist + 1);
              __pyx_t_10.n_subs = (__pyx_v_cand.n_subs + 1);

              /* "fuzzysearch/_generic_search.pyx":155
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs + 1,
 *                                 cand.n_ins, cand.n_dels,             # <<<<<<<<<<<<<<
 *                             )
 *                             n_new_candidates += 1
 */
              __pyx_t_10.n_ins = __pyx_v_cand.n_ins;
              __pyx_t_10.n_dels = __pyx_v_cand.n_dels;

              /* "fuzzysearch/_generic_search.pyx":152
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs + 1,
 */
              (__pyx_v_new_candidates[__pyx_v_n_new_candidates]) = __pyx_t_10;

              /* "fuzzysearch/_generic_search.pyx":157
 *                                 cand.n_ins, cand.n_dels,
 *                             )
 *                             n_new_candidates += 1             # <<<<<<<<<<<<<<
 *                         elif cand.n_dels < max_deletions and cand.n_ins < max_insertions:
 *                             # add a candidate skipping both a sequence char and a
 */
              __pyx_v_n_new_candidates = (__pyx_v_n_new_candidates + 1);

              /* "fuzzysearch/_generic_search.pyx":149
 * 
 *                     if cand.subseq_index + 1 < subseq_len:
 *                         if cand.n_subs < max_substitutions:             # <<<<<<<<<<<<<<
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 */
              goto __pyx_L19;
            }

            /* "fuzzysearch/_generic_search.pyx":158
 *                             )
 *                             n_new_candidates += 1
 *                         elif cand.n_dels < max_deletions and cand.n_ins < max_insertions:             # <<<<<<<<<<<<<<
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 */
            __pyx_t_20 = ((__pyx_v_cand.n_dels < __pyx_v_max_deletions) != 0);
            if (__pyx_t_20) {
            } else {
              __pyx_t_4 = __pyx_t_20;
              goto __pyx_L20_bool_binop_done;
            }
            __pyx_t_20 = ((__pyx_v_cand.n_ins < __pyx_v_max_insertions) != 0);
            __pyx_t_4 = __pyx_t_20;
            __pyx_L20_bool_binop_done:;
            if (__pyx_t_4) {

              /* "fuzzysearch/_generic_search.pyx":162
 *                             # subsequence char
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                 cand.start, cand.subseq_index + 1,             # <<<<<<<<<<<<<<
 *                                 cand.l_dist + 1, cand.n_subs,
 *                                 cand.n_ins + 1, cand.n_dels + 1,
 */
              __pyx_t_10.start = __pyx_v_cand.start;
              __pyx_t_10.subseq_index = (__pyx_v_cand.subseq_index + 1);

              /* "fuzzysearch/_generic_search.pyx":163
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs,             # <<<<<<<<<<<<<<
 *                                 cand.n_ins + 1, cand.n_dels + 1,
 *                             )
 */
              __pyx_t_10.l_dist = (__pyx_v_cand.l_dist + 1);
              __pyx_t_10.n_subs = __pyx_v_cand.n_subs;

              /* "fuzzysearch/_generic_search.pyx":164
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs,
 *                                 cand.n_ins + 1, cand.n_dels + 1,             # <<<<<<<<<<<<<<
 *                             )
 *                             n_new_candidates += 1
 */
              __pyx_t_10.n_ins = (__pyx_v_cand.n_ins + 1);
              __pyx_t_10.n_dels = (__pyx_v_cand.n_dels + 1);

              /* "fuzzysearch/_generic_search.pyx":161
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 *                             new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                                 cand.start, cand.subseq_index + 1,
 *                                 cand.l_dist + 1, cand.n_subs,
 */
              (__pyx_v_new_candidates[__pyx_v_n_new_candidates]) = __pyx_t_10;

              /* "fuzzysearch/_generic_search.pyx":166
 *                                 cand.n_ins + 1, cand.n_dels + 1,
 *                             )
 *                             n_new_candidates += 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         # cand.subseq_index == _subseq_len - 1
 */
              __pyx_v_n_new_candidates = (__pyx_v_n_new_candidates + 1);

              /* "fuzzysearch/_generic_search.pyx":158
 *                             )
 *                             n_new_candidates += 1
 *                         elif cand.n_dels < max_deletions and cand.n_ins < max_insertions:             # <<<<<<<<<<<<<<
 *                             # add a candidate skipping both a sequence char and a
 *                             # subsequence char
 */
            }
            __pyx_L19:;

            /* "fuzzysearch/_generic_search.pyx":148
 *                         n_new_candidates += 1
 * 
 *                     if cand.subseq_index + 1 < subseq_len:             # <<<<<<<<<<<<<<
 *                         if cand.n_subs < max_substitutions:
 *                             # add a candidate skipping both a sequence char and a
 */
            goto __pyx_L18;
          }

          /* "fuzzysearch/_generic_search.pyx":169
 *                     else:
 *                         # cand.subseq_index == _subseq_len - 1
 *                         if (             # <<<<<<<<<<<<<<
 *                                 cand.n_subs < max_substitutions or
 *                                 (
 */
          /*else*/ {

            /* "fuzzysearch/_generic_search.pyx":170
 *                         # cand.subseq_index == _subseq_len - 1
 *                         if (
 *                                 cand.n_subs < max_substitutions or             # <<<<<<<<<<<<<<
 *                                 (
 *                                     cand.n_dels < max_deletions and
 */
            __pyx_t_20 = ((__pyx_v_cand.n_subs < __pyx_v_max_substitutions) != 0);
            if (!__pyx_t_20) {
            } else {
              __pyx_t_4 = __pyx_t_20;
              goto __pyx_L23_bool_binop_done;
            }

            /* "fuzzysearch/_generic_search.pyx":172
 *                                 cand.n_subs < max_substitutions or
 *                                 (
 *                                     cand.n_dels < max_deletions and             # <<<<<<<<<<<<<<
 *                                     cand.n_ins < max_insertions
 *                                 )
 */
            __pyx_t_20 = ((__pyx_v_cand.n_dels < __pyx_v_max_deletions) != 0);
            if (__pyx_t_20) {
            } else {
              __pyx_t_4 = __pyx_t_20;
              goto __pyx_L23_bool_binop_done;
            }

            /* "fuzzysearch/_generic_search.pyx":173
 *                                 (
 *                                     cand.n_dels < max_deletions and
 *                                     cand.n_ins < max_insertions             # <<<<<<<<<<<<<<
 *                                 )
 *                         ):
 */
            __pyx_t_20 = ((__pyx_v_cand.n_ins < __pyx_v_max_insertions) != 0);
            __pyx_t_4 = __pyx_t_20;
            __pyx_L23_bool_binop_done:;

            /* "fuzzysearch/_generic_search.pyx":169
 *                     else:
 *                         # cand.subseq_index == _subseq_len - 1
 *                         if (             # <<<<<<<<<<<<<<
 *                                 cand.n_subs < max_substitutions or
 *                                 (
 */
            if (__pyx_t_4) {

              /* "fuzzysearch/_generic_search.pyx":176
 *                                 )
 *                         ):
 *                             matches.append(Match(cand.start, index + 1, cand.l_dist + 1))             # <<<<<<<<<<<<<<
 * 
 *                     # try skipping subsequence chars
 */
              __pyx_t_12 = __Pyx_GetModuleGlobalName(__pyx_n_s_Match); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 176, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_12);
              __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_cand.start); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 176, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_15 = __Pyx_PyInt_FromSize_t((__pyx_v_index + 1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 176, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_15);
              __pyx_t_14 = __Pyx_PyInt_From_long((__pyx_v_cand.l_dist + 1)); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 176, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_13 = NULL;
              __pyx_t_17 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_12))) {
                __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_12);
                if (likely(__pyx_t_13)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_12);
                  __Pyx_INCREF(__pyx_t_13);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_12, function);
                  __pyx_t_17 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_12)) {
                PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_18, __pyx_t_15, __pyx_t_14};
                __pyx_t_11 = __Pyx_PyFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 176, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_12)) {
                PyObject *__pyx_temp[4] = {__pyx_t_13, __pyx_t_18, __pyx_t_15, __pyx_t_14};
                __pyx_t_11 = __Pyx_PyCFunction_FastCall(__pyx_t_12, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 176, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              } else
              #endif
              {
                __pyx_t_16 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 176, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_16);
                if (__pyx_t_13) {
                  __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_13); __pyx_t_13 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_18);
                PyTuple_SET_ITEM(__pyx_t_16, 0+__pyx_t_17, __pyx_t_18);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_16, 1+__pyx_t_17, __pyx_t_15);
                __Pyx_GIVEREF(__pyx_t_14);
                PyTuple_SET_ITEM(__pyx_t_16, 2+__pyx_t_17, __pyx_t_14);
                __pyx_t_18 = 0;
                __pyx_t_15 = 0;
                __pyx_t_14 = 0;
                __pyx_t_11 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_16, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 176, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_11);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
              }
              __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
              __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_11); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 176, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "fuzzysearch/_generic_search.pyx":169
 *                     else:
 *                         # cand.subseq_index == _subseq_len - 1
 *                         if (             # <<<<<<<<<<<<<<
 *                                 cand.n_subs < max_substitutions or
 *                                 (
 */
            }
          }
          __pyx_L18:;

          /* "fuzzysearch/_generic_search.pyx":179
 * 
 *                     # try skipping subsequence chars
 *                     for n_skipped in xrange(1, min(max_deletions - cand.n_dels, max_l_dist - cand.l_dist) + 1):             # <<<<<<<<<<<<<<
 *                         # if skipping n_dels sub-sequence chars reaches the end
 *                         # of the sub-sequence, yield a match
 */
          __pyx_t_21 = (__pyx_v_max_l_dist - __pyx_v_cand.l_dist);
          __pyx_t_22 = (__pyx_v_max_deletions - __pyx_v_cand.n_dels);
          if (((__pyx_t_21 < __pyx_t_22) != 0)) {
            __pyx_t_23 = __pyx_t_21;
          } else {
            __pyx_t_23 = __pyx_t_22;
          }
          __pyx_t_11 = __Pyx_PyInt_From_long((__pyx_t_23 + 1)); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 179, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_11);
          __pyx_t_12 = PyTuple_New(2); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 179, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_INCREF(__pyx_int_1);
          __Pyx_GIVEREF(__pyx_int_1);
          PyTuple_SET_ITEM(__pyx_t_12, 0, __pyx_int_1);
          __Pyx_GIVEREF(__pyx_t_11);
          PyTuple_SET_ITEM(__pyx_t_12, 1, __pyx_t_11);
          __pyx_t_11 = 0;
          __pyx_t_11 = __Pyx_PyObject_Call(__pyx_builtin_xrange, __pyx_t_12, NULL); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 179, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_11);
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (likely(PyList_CheckExact(__pyx_t_11)) || PyTuple_CheckExact(__pyx_t_11)) {
            __pyx_t_12 = __pyx_t_11; __Pyx_INCREF(__pyx_t_12); __pyx_t_24 = 0;
            __pyx_t_25 = NULL;
          } else {
            __pyx_t_24 = -1; __pyx_t_12 = PyObject_GetIter(__pyx_t_11); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 179, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_12);
            __pyx_t_25 = Py_TYPE(__pyx_t_12)->tp_iternext; if (unlikely(!__pyx_t_25)) __PYX_ERR(0, 179, __pyx_L6_error)
          }
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          for (;;) {
            if (likely(!__pyx_t_25)) {
              if (likely(PyList_CheckExact(__pyx_t_12))) {
                if (__pyx_t_24 >= PyList_GET_SIZE(__pyx_t_12)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_11 = PyList_GET_ITEM(__pyx_t_12, __pyx_t_24); __Pyx_INCREF(__pyx_t_11); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L6_error)
                #else
                __pyx_t_11 = PySequence_ITEM(__pyx_t_12, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 179, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_11);
                #endif
              } else {
                if (__pyx_t_24 >= PyTuple_GET_SIZE(__pyx_t_12)) break;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                __pyx_t_11 = PyTuple_GET_ITEM(__pyx_t_12, __pyx_t_24); __Pyx_INCREF(__pyx_t_11); __pyx_t_24++; if (unlikely(0 < 0)) __PYX_ERR(0, 179, __pyx_L6_error)
                #else
                __pyx_t_11 = PySequence_ITEM(__pyx_t_12, __pyx_t_24); __pyx_t_24++; if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 179, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_11);
                #endif
              }
            } else {
              __pyx_t_11 = __pyx_t_25(__pyx_t_12);
              if (unlikely(!__pyx_t_11)) {
                PyObject* exc_type = PyErr_Occurred();
                if (exc_type) {
                  if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                  else __PYX_ERR(0, 179, __pyx_L6_error)
                }
                break;
              }
              __Pyx_GOTREF(__pyx_t_11);
            }
            __Pyx_XDECREF_SET(__pyx_v_n_skipped, __pyx_t_11);
            __pyx_t_11 = 0;

            /* "fuzzysearch/_generic_search.pyx":182
 *                         # if skipping n_dels sub-sequence chars reaches the end
 *                         # of the sub-sequence, yield a match
 *                         if cand.subseq_index + n_skipped == subseq_len:             # <<<<<<<<<<<<<<
 *                             matches.append(Match(cand.start, index + 1,
 *                                                  cand.l_dist + n_skipped))
 */
            __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cand.subseq_index); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 182, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_16 = PyNumber_Add(__pyx_t_11, __pyx_v_n_skipped); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 182, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_16);
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_11 = __Pyx_PyInt_FromSize_t(__pyx_v_subseq_len); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 182, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_11);
            __pyx_t_14 = PyObject_RichCompare(__pyx_t_16, __pyx_t_11, Py_EQ); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 182, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 182, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            if (__pyx_t_4) {

              /* "fuzzysearch/_generic_search.pyx":183
 *                         # of the sub-sequence, yield a match
 *                         if cand.subseq_index + n_skipped == subseq_len:
 *                             matches.append(Match(cand.start, index + 1,             # <<<<<<<<<<<<<<
 *                                                  cand.l_dist + n_skipped))
 *                             break
 */
              __pyx_t_11 = __Pyx_GetModuleGlobalName(__pyx_n_s_Match); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 183, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_cand.start); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 183, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_16);
              __pyx_t_15 = __Pyx_PyInt_FromSize_t((__pyx_v_index + 1)); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 183, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_15);

              /* "fuzzysearch/_generic_search.pyx":184
 *                         if cand.subseq_index + n_skipped == subseq_len:
 *                             matches.append(Match(cand.start, index + 1,
 *                                                  cand.l_dist + n_skipped))             # <<<<<<<<<<<<<<
 *                             break
 *                         # otherwise, if skipping n_skipped sub-sequence chars
 */
              __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 184, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_18);
              __pyx_t_13 = PyNumber_Add(__pyx_t_18, __pyx_v_n_skipped); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 184, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_13);
              __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
              __pyx_t_18 = NULL;
              __pyx_t_17 = 0;
              if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_11))) {
                __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_11);
                if (likely(__pyx_t_18)) {
                  PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
                  __Pyx_INCREF(__pyx_t_18);
                  __Pyx_INCREF(function);
                  __Pyx_DECREF_SET(__pyx_t_11, function);
                  __pyx_t_17 = 1;
                }
              }
              #if CYTHON_FAST_PYCALL
              if (PyFunction_Check(__pyx_t_11)) {
                PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_16, __pyx_t_15, __pyx_t_13};
                __pyx_t_14 = __Pyx_PyFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 183, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              } else
              #endif
              #if CYTHON_FAST_PYCCALL
              if (__Pyx_PyFastCFunction_Check(__pyx_t_11)) {
                PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_16, __pyx_t_15, __pyx_t_13};
                __pyx_t_14 = __Pyx_PyCFunction_FastCall(__pyx_t_11, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 183, __pyx_L6_error)
                __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
              } else
              #endif
              {
                __pyx_t_26 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 183, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_26);
                if (__pyx_t_18) {
                  __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_26, 0, __pyx_t_18); __pyx_t_18 = NULL;
                }
                __Pyx_GIVEREF(__pyx_t_16);
                PyTuple_SET_ITEM(__pyx_t_26, 0+__pyx_t_17, __pyx_t_16);
                __Pyx_GIVEREF(__pyx_t_15);
                PyTuple_SET_ITEM(__pyx_t_26, 1+__pyx_t_17, __pyx_t_15);
                __Pyx_GIVEREF(__pyx_t_13);
                PyTuple_SET_ITEM(__pyx_t_26, 2+__pyx_t_17, __pyx_t_13);
                __pyx_t_16 = 0;
                __pyx_t_15 = 0;
                __pyx_t_13 = 0;
                __pyx_t_14 = __Pyx_PyObject_Call(__pyx_t_11, __pyx_t_26, NULL); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 183, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
              }
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

              /* "fuzzysearch/_generic_search.pyx":183
 *                         # of the sub-sequence, yield a match
 *                         if cand.subseq_index + n_skipped == subseq_len:
 *                             matches.append(Match(cand.start, index + 1,             # <<<<<<<<<<<<<<
 *                                                  cand.l_dist + n_skipped))
 *                             break
 */
              __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_14); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 183, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

              /* "fuzzysearch/_generic_search.pyx":185
 *                             matches.append(Match(cand.start, index + 1,
 *                                                  cand.l_dist + n_skipped))
 *                             break             # <<<<<<<<<<<<<<
 *                         # otherwise, if skipping n_skipped sub-sequence chars
 *                         # reaches a sub-sequence char identical to this sequence
 */
              goto __pyx_L27_break;

              /* "fuzzysearch/_generic_search.pyx":182
 *                         # if skipping n_dels sub-sequence chars reaches the end
 *                         # of the sub-sequence, yield a match
 *                         if cand.subseq_index + n_skipped == subseq_len:             # <<<<<<<<<<<<<<
 *                             matches.append(Match(cand.start, index + 1,
 *                                                  cand.l_dist + n_skipped))
 */
            }

            /* "fuzzysearch/_generic_search.pyx":189
 *                         # reaches a sub-sequence char identical to this sequence
 *                         # char ...
 *                         elif seq_char == subsequence[cand.subseq_index + n_skipped]:             # <<<<<<<<<<<<<<
 *                             # if this is the last char of the sub-sequence, yield
 *                             # a match
 */
            __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_cand.subseq_index); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 189, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_14);
            __pyx_t_11 = PyNumber_Add(__pyx_t_14, __pyx_v_n_skipped); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 189, __pyx_L6_error)
            __Pyx_GOTREF(__pyx_t_11);
            __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
            __pyx_t_27 = __Pyx_PyIndex_AsSsize_t(__pyx_t_11); if (unlikely((__pyx_t_27 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 189, __pyx_L6_error)
            __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
            __pyx_t_4 = ((__pyx_v_seq_char == (__pyx_v_subsequence[__pyx_t_27])) != 0);
            if (__pyx_t_4) {

              /* "fuzzysearch/_generic_search.pyx":192
 *                             # if this is the last char of the sub-sequence, yield
 *                             # a match
 *                             if cand.subseq_index + n_skipped + 1 == subseq_len:             # <<<<<<<<<<<<<<
 *                                 matches.append(Match(cand.start, index + 1,
 *                                                      cand.l_dist + n_skipped))
 */
              __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cand.subseq_index); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_11);
              __pyx_t_14 = PyNumber_Add(__pyx_t_11, __pyx_v_n_skipped); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_14);
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __pyx_t_11 = __Pyx_PyInt_AddObjC(__pyx_t_14, __pyx_int_1, 1, 0); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_11);
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_14 = __Pyx_PyInt_FromSize_t(__pyx_v_subseq_len); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_GOTREF(__pyx_t_14);
              __pyx_t_26 = PyObject_RichCompare(__pyx_t_11, __pyx_t_14, Py_EQ); __Pyx_XGOTREF(__pyx_t_26); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
              __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
              __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_26); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 192, __pyx_L6_error)
              __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
              if (__pyx_t_4) {

                /* "fuzzysearch/_generic_search.pyx":193
 *                             # a match
 *                             if cand.subseq_index + n_skipped + 1 == subseq_len:
 *                                 matches.append(Match(cand.start, index + 1,             # <<<<<<<<<<<<<<
 *                                                      cand.l_dist + n_skipped))
 *                             # otherwise add a candidate skipping n_skipped
 */
                __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_Match); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 193, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_cand.start); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 193, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_11);
                __pyx_t_13 = __Pyx_PyInt_FromSize_t((__pyx_v_index + 1)); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 193, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_13);

                /* "fuzzysearch/_generic_search.pyx":194
 *                             if cand.subseq_index + n_skipped + 1 == subseq_len:
 *                                 matches.append(Match(cand.start, index + 1,
 *                                                      cand.l_dist + n_skipped))             # <<<<<<<<<<<<<<
 *                             # otherwise add a candidate skipping n_skipped
 *                             # subsequence chars
 */
                __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 194, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_15);
                __pyx_t_16 = PyNumber_Add(__pyx_t_15, __pyx_v_n_skipped); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 194, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_16);
                __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
                __pyx_t_15 = NULL;
                __pyx_t_17 = 0;
                if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
                  __pyx_t_15 = PyMethod_GET_SELF(__pyx_t_14);
                  if (likely(__pyx_t_15)) {
                    PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
                    __Pyx_INCREF(__pyx_t_15);
                    __Pyx_INCREF(function);
                    __Pyx_DECREF_SET(__pyx_t_14, function);
                    __pyx_t_17 = 1;
                  }
                }
                #if CYTHON_FAST_PYCALL
                if (PyFunction_Check(__pyx_t_14)) {
                  PyObject *__pyx_temp[4] = {__pyx_t_15, __pyx_t_11, __pyx_t_13, __pyx_t_16};
                  __pyx_t_26 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 193, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  __Pyx_GOTREF(__pyx_t_26);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                } else
                #endif
                #if CYTHON_FAST_PYCCALL
                if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
                  PyObject *__pyx_temp[4] = {__pyx_t_15, __pyx_t_11, __pyx_t_13, __pyx_t_16};
                  __pyx_t_26 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 193, __pyx_L6_error)
                  __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
                  __Pyx_GOTREF(__pyx_t_26);
                  __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
                  __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
                  __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
                } else
                #endif
                {
                  __pyx_t_18 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 193, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_18);
                  if (__pyx_t_15) {
                    __Pyx_GIVEREF(__pyx_t_15); PyTuple_SET_ITEM(__pyx_t_18, 0, __pyx_t_15); __pyx_t_15 = NULL;
                  }
                  __Pyx_GIVEREF(__pyx_t_11);
                  PyTuple_SET_ITEM(__pyx_t_18, 0+__pyx_t_17, __pyx_t_11);
                  __Pyx_GIVEREF(__pyx_t_13);
                  PyTuple_SET_ITEM(__pyx_t_18, 1+__pyx_t_17, __pyx_t_13);
                  __Pyx_GIVEREF(__pyx_t_16);
                  PyTuple_SET_ITEM(__pyx_t_18, 2+__pyx_t_17, __pyx_t_16);
                  __pyx_t_11 = 0;
                  __pyx_t_13 = 0;
                  __pyx_t_16 = 0;
                  __pyx_t_26 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_18, NULL); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 193, __pyx_L6_error)
                  __Pyx_GOTREF(__pyx_t_26);
                  __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
                }
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;

                /* "fuzzysearch/_generic_search.pyx":193
 *                             # a match
 *                             if cand.subseq_index + n_skipped + 1 == subseq_len:
 *                                 matches.append(Match(cand.start, index + 1,             # <<<<<<<<<<<<<<
 *                                                      cand.l_dist + n_skipped))
 *                             # otherwise add a candidate skipping n_skipped
 */
                __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_26); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 193, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;

                /* "fuzzysearch/_generic_search.pyx":192
 *                             # if this is the last char of the sub-sequence, yield
 *                             # a match
 *                             if cand.subseq_index + n_skipped + 1 == subseq_len:             # <<<<<<<<<<<<<<
 *                                 matches.append(Match(cand.start, index + 1,
 *                                                      cand.l_dist + n_skipped))
 */
                goto __pyx_L29;
              }

              /* "fuzzysearch/_generic_search.pyx":198
 *                             # subsequence chars
 *                             else:
 *                                 new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                                     cand.start, cand.subseq_index + 1 + n_skipped,
 *                                     cand.l_dist + n_skipped, cand.n_subs,
 */
              /*else*/ {

                /* "fuzzysearch/_generic_search.pyx":199
 *                             else:
 *                                 new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                     cand.start, cand.subseq_index + 1 + n_skipped,             # <<<<<<<<<<<<<<
 *                                     cand.l_dist + n_skipped, cand.n_subs,
 *                                     cand.n_ins, cand.n_dels + n_skipped,
 */
                __pyx_t_10.start = __pyx_v_cand.start;
                __pyx_t_26 = __Pyx_PyInt_From_long((__pyx_v_cand.subseq_index + 1)); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 199, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_26);
                __pyx_t_14 = PyNumber_Add(__pyx_t_26, __pyx_v_n_skipped); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 199, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
                __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_14); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 199, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __pyx_t_10.subseq_index = __pyx_t_17;

                /* "fuzzysearch/_generic_search.pyx":200
 *                                 new_candidates[n_new_candidates] = GenericSearchCandidate(
 *                                     cand.start, cand.subseq_index + 1 + n_skipped,
 *                                     cand.l_dist + n_skipped, cand.n_subs,             # <<<<<<<<<<<<<<
 *                                     cand.n_ins, cand.n_dels + n_skipped,
 *                                 )
 */
                __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 200, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_14);
                __pyx_t_26 = PyNumber_Add(__pyx_t_14, __pyx_v_n_skipped); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 200, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_26);
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_26); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 200, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
                __pyx_t_10.l_dist = __pyx_t_17;
                __pyx_t_10.n_subs = __pyx_v_cand.n_subs;

                /* "fuzzysearch/_generic_search.pyx":201
 *                                     cand.start, cand.subseq_index + 1 + n_skipped,
 *                                     cand.l_dist + n_skipped, cand.n_subs,
 *                                     cand.n_ins, cand.n_dels + n_skipped,             # <<<<<<<<<<<<<<
 *                                 )
 *                                 n_new_candidates += 1
 */
                __pyx_t_10.n_ins = __pyx_v_cand.n_ins;
                __pyx_t_26 = __Pyx_PyInt_From_int(__pyx_v_cand.n_dels); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 201, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_26);
                __pyx_t_14 = PyNumber_Add(__pyx_t_26, __pyx_v_n_skipped); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 201, __pyx_L6_error)
                __Pyx_GOTREF(__pyx_t_14);
                __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
                __pyx_t_17 = __Pyx_PyInt_As_int(__pyx_t_14); if (unlikely((__pyx_t_17 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 201, __pyx_L6_error)
                __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
                __pyx_t_10.n_dels = __pyx_t_17;

                /* "fuzzysearch/_generic_search.pyx":198
 *                             # subsequence chars
 *                             else:
 *                                 new_candidates[n_new_candidates] = GenericSearchCandidate(             # <<<<<<<<<<<<<<
 *                                     cand.start, cand.subseq_index + 1 + n_skipped,
 *                                     cand.l_dist + n_skipped, cand.n_subs,
 */
                (__pyx_v_new_candidates[__pyx_v_n_new_candidates]) = __pyx_t_10;

                /* "fuzzysearch/_generic_search.pyx":203
 *                                     cand.n_ins, cand.n_dels + n_skipped,
 *                                 )
 *                                 n_new_candidates += 1             # <<<<<<<<<<<<<<
 *                             break
 *                     # note: if the above loop ends without a break, that means that
 */
                __pyx_v_n_new_candidates = (__pyx_v_n_new_candidates + 1);
              }
              __pyx_L29:;

              /* "fuzzysearch/_generic_search.pyx":204
 *                                 )
 *                                 n_new_candidates += 1
 *                             break             # <<<<<<<<<<<<<<
 *                     # note: if the above loop ends without a break, that means that
 *                     # no candidate could be added / yielded by skipping sub-sequence
 */
              goto __pyx_L27_break;

              /* "fuzzysearch/_generic_search.pyx":189
 *                         # reaches a sub-sequence char identical to this sequence
 *                         # char ...
 *                         elif seq_char == subsequence[cand.subseq_index + n_skipped]:             # <<<<<<<<<<<<<<
 *                             # if this is the last char of the sub-sequence, yield
 *                             # a match
 */
            }

            /* "fuzzysearch/_generic_search.pyx":179
 * 
 *                     # try skipping subsequence chars
 *                     for n_skipped in xrange(1, min(max_deletions - cand.n_dels, max_l_dist - cand.l_dist) + 1):             # <<<<<<<<<<<<<<
 *                         # if skipping n_dels sub-sequence chars reaches the end
 *                         # of the sub-sequence, yield a match
 */
          }
          __pyx_L27_break:;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        }
        __pyx_L14:;
        __pyx_L10_continue:;
      }

      /* "fuzzysearch/_generic_search.pyx":210
 * 
 *             # new_candidates = candidates; candidates = []
 *             _tmp = candidates             # <<<<<<<<<<<<<<
 *             candidates = new_candidates
 *             new_candidates = _tmp
 */
      __pyx_v__tmp = __pyx_v_candidates;

      /* "fuzzysearch/_generic_search.pyx":211
 *             # new_candidates = candidates; candidates = []
 *             _tmp = candidates
 *             candidates = new_candidates             # <<<<<<<<<<<<<<
 *             new_candidates = _tmp
 *             n_candidates = n_new_candidates
 */
      __pyx_v_candidates = __pyx_v_new_candidates;

      /* "fuzzysearch/_generic_search.pyx":212
 *             _tmp = candidates
 *             candidates = new_candidates
 *             new_candidates = _tmp             # <<<<<<<<<<<<<<
 *             n_candidates = n_new_candidates
 *             n_new_candidates = 0
 */
      __pyx_v_new_candidates = __pyx_v__tmp;

      /* "fuzzysearch/_generic_search.pyx":213
 *             candidates = new_candidates
 *             new_candidates = _tmp
 *             n_candidates = n_new_candidates             # <<<<<<<<<<<<<<
 *             n_new_candidates = 0
 * 
 */
      __pyx_v_n_candidates = __pyx_v_n_new_candidates;

      /* "fuzzysearch/_generic_search.pyx":214
 *             new_candidates = _tmp
 *             n_candidates = n_new_candidates
 *             n_new_candidates = 0             # <<<<<<<<<<<<<<
 * 
 *             if have_realloced:
 */
      __pyx_v_n_new_candidates = 0;

      /* "fuzzysearch/_generic_search.pyx":216
 *             n_new_candidates = 0
 * 
 *             if have_realloced:             # <<<<<<<<<<<<<<
 *                 have_realloced = False
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 */
      __pyx_t_4 = (__pyx_v_have_realloced != 0);
      if (__pyx_t_4) {

        /* "fuzzysearch/_generic_search.pyx":217
 * 
 *             if have_realloced:
 *                 have_realloced = False             # <<<<<<<<<<<<<<
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                 if _tmp is NULL:
 */
        __pyx_v_have_realloced = 0;

        /* "fuzzysearch/_generic_search.pyx":218
 *             if have_realloced:
 *                 have_realloced = False
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))             # <<<<<<<<<<<<<<
 *                 if _tmp is NULL:
 *                     raise MemoryError()
 */
        __pyx_v__tmp = ((struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate *)realloc(__pyx_v_new_candidates, (__pyx_v_alloc_size * (sizeof(struct __pyx_t_11fuzzysearch_15_generic_search_GenericSearchCandidate)))));

        /* "fuzzysearch/_generic_search.pyx":219
 *                 have_realloced = False
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                 if _tmp is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 *                 new_candidates = _tmp
 */
        __pyx_t_4 = ((__pyx_v__tmp == NULL) != 0);
        if (__pyx_t_4) {

          /* "fuzzysearch/_generic_search.pyx":220
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                 if _tmp is NULL:
 *                     raise MemoryError()             # <<<<<<<<<<<<<<
 *                 new_candidates = _tmp
 * 
 */
          PyErr_NoMemory(); __PYX_ERR(0, 220, __pyx_L6_error)

          /* "fuzzysearch/_generic_search.pyx":219
 *                 have_realloced = False
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 *                 if _tmp is NULL:             # <<<<<<<<<<<<<<
 *                     raise MemoryError()
 *                 new_candidates = _tmp
 */
        }

        /* "fuzzysearch/_generic_search.pyx":221
 *                 if _tmp is NULL:
 *                     raise MemoryError()
 *                 new_candidates = _tmp             # <<<<<<<<<<<<<<
 * 
 *             index += 1
 */
        __pyx_v_new_candidates = __pyx_v__tmp;

        /* "fuzzysearch/_generic_search.pyx":216
 *             n_new_candidates = 0
 * 
 *             if have_realloced:             # <<<<<<<<<<<<<<
 *                 have_realloced = False
 *                 _tmp = <GenericSearchCandidate *>realloc(new_candidates, alloc_size * sizeof(GenericSearchCandidate))
 */
      }

      /* "fuzzysearch/_generic_search.pyx":223
 *                 new_candidates = _tmp
 * 
 *             index += 1             # <<<<<<<<<<<<<<
 * 
 *         for n_cand in xrange(n_candidates):
 */
      __pyx_v_index = (__pyx_v_index + 1);
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "fuzzysearch/_generic_search.pyx":225
 *             index += 1
 * 
 *         for n_cand in xrange(n_candidates):             # <<<<<<<<<<<<<<
 *             cand = candidates[n_cand]
 *             # note: index == length(sequence)
 */
    __pyx_t_3 = __pyx_v_n_candidates;
    for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_3; __pyx_t_1+=1) {
      __pyx_v_n_cand = __pyx_t_1;

      /* "fuzzysearch/_generic_search.pyx":226
 * 
 *         for n_cand in xrange(n_candidates):
 *             cand = candidates[n_cand]             # <<<<<<<<<<<<<<
 *             # note: index == length(sequence)
 *             n_skipped = subseq_len - cand.subseq_index
 */
      __pyx_v_cand = (__pyx_v_candidates[__pyx_v_n_cand]);

      /* "fuzzysearch/_generic_search.pyx":228
 *             cand = candidates[n_cand]
 *             # note: index == length(sequence)
 *             n_skipped = subseq_len - cand.subseq_index             # <<<<<<<<<<<<<<
 *             if cand.n_dels + n_skipped <= max_deletions and \
 *                cand.l_dist + n_skipped <= max_l_dist:
 */
      __pyx_t_5 = __Pyx_PyInt_FromSize_t((__pyx_v_subseq_len - __pyx_v_cand.subseq_index)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 228, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_n_skipped, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "fuzzysearch/_generic_search.pyx":229
 *             # note: index == length(sequence)
 *             n_skipped = subseq_len - cand.subseq_index
 *             if cand.n_dels + n_skipped <= max_deletions and \             # <<<<<<<<<<<<<<
 *                cand.l_dist + n_skipped <= max_l_dist:
 *                 matches.append(Match(cand.start, index, cand.l_dist + n_skipped))
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cand.n_dels); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 229, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_12 = PyNumber_Add(__pyx_t_5, __pyx_v_n_skipped); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 229, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_12);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_deletions); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 229, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_14 = PyObject_RichCompare(__pyx_t_12, __pyx_t_5, Py_LE); __Pyx_XGOTREF(__pyx_t_14); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 229, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_14); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 229, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      if (__pyx_t_20) {
      } else {
        __pyx_t_4 = __pyx_t_20;
        goto __pyx_L35_bool_binop_done;
      }

      /* "fuzzysearch/_generic_search.pyx":230
 *             n_skipped = subseq_len - cand.subseq_index
 *             if cand.n_dels + n_skipped <= max_deletions and \
 *                cand.l_dist + n_skipped <= max_l_dist:             # <<<<<<<<<<<<<<
 *                 matches.append(Match(cand.start, index, cand.l_dist + n_skipped))
 * 
 */
      __pyx_t_14 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 230, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_5 = PyNumber_Add(__pyx_t_14, __pyx_v_n_skipped); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 230, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_14 = __Pyx_PyInt_From_unsigned_int(__pyx_v_max_l_dist); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 230, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_14);
      __pyx_t_12 = PyObject_RichCompare(__pyx_t_5, __pyx_t_14, Py_LE); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 230, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
      __pyx_t_20 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_20 < 0)) __PYX_ERR(0, 230, __pyx_L6_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __pyx_t_4 = __pyx_t_20;
      __pyx_L35_bool_binop_done:;

      /* "fuzzysearch/_generic_search.pyx":229
 *             # note: index == length(sequence)
 *             n_skipped = subseq_len - cand.subseq_index
 *             if cand.n_dels + n_skipped <= max_deletions and \             # <<<<<<<<<<<<<<
 *                cand.l_dist + n_skipped <= max_l_dist:
 *                 matches.append(Match(cand.start, index, cand.l_dist + n_skipped))
 */
      if (__pyx_t_4) {

        /* "fuzzysearch/_generic_search.pyx":231
 *             if cand.n_dels + n_skipped <= max_deletions and \
 *                cand.l_dist + n_skipped <= max_l_dist:
 *                 matches.append(Match(cand.start, index, cand.l_dist + n_skipped))             # <<<<<<<<<<<<<<
 * 
 *     finally:
 */
        __pyx_t_14 = __Pyx_GetModuleGlobalName(__pyx_n_s_Match); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_14);
        __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cand.start); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_26 = __Pyx_PyInt_FromSize_t(__pyx_v_index); if (unlikely(!__pyx_t_26)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_26);
        __pyx_t_18 = __Pyx_PyInt_From_int(__pyx_v_cand.l_dist); if (unlikely(!__pyx_t_18)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_18);
        __pyx_t_16 = PyNumber_Add(__pyx_t_18, __pyx_v_n_skipped); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_GOTREF(__pyx_t_16);
        __Pyx_DECREF(__pyx_t_18); __pyx_t_18 = 0;
        __pyx_t_18 = NULL;
        __pyx_t_17 = 0;
        if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_14))) {
          __pyx_t_18 = PyMethod_GET_SELF(__pyx_t_14);
          if (likely(__pyx_t_18)) {
            PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_14);
            __Pyx_INCREF(__pyx_t_18);
            __Pyx_INCREF(function);
            __Pyx_DECREF_SET(__pyx_t_14, function);
            __pyx_t_17 = 1;
          }
        }
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_14)) {
          PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_5, __pyx_t_26, __pyx_t_16};
          __pyx_t_12 = __Pyx_PyFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 231, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_14)) {
          PyObject *__pyx_temp[4] = {__pyx_t_18, __pyx_t_5, __pyx_t_26, __pyx_t_16};
          __pyx_t_12 = __Pyx_PyCFunction_FastCall(__pyx_t_14, __pyx_temp+1-__pyx_t_17, 3+__pyx_t_17); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 231, __pyx_L6_error)
          __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_26); __pyx_t_26 = 0;
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        } else
        #endif
        {
          __pyx_t_13 = PyTuple_New(3+__pyx_t_17); if (unlikely(!__pyx_t_13)) __PYX_ERR(0, 231, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_13);
          if (__pyx_t_18) {
            __Pyx_GIVEREF(__pyx_t_18); PyTuple_SET_ITEM(__pyx_t_13, 0, __pyx_t_18); __pyx_t_18 = NULL;
          }
          __Pyx_GIVEREF(__pyx_t_5);
          PyTuple_SET_ITEM(__pyx_t_13, 0+__pyx_t_17, __pyx_t_5);
          __Pyx_GIVEREF(__pyx_t_26);
          PyTuple_SET_ITEM(__pyx_t_13, 1+__pyx_t_17, __pyx_t_26);
          __Pyx_GIVEREF(__pyx_t_16);
          PyTuple_SET_ITEM(__pyx_t_13, 2+__pyx_t_17, __pyx_t_16);
          __pyx_t_5 = 0;
          __pyx_t_26 = 0;
          __pyx_t_16 = 0;
          __pyx_t_12 = __Pyx_PyObject_Call(__pyx_t_14, __pyx_t_13, NULL); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 231, __pyx_L6_error)
          __Pyx_GOTREF(__pyx_t_12);
          __Pyx_DECREF(__pyx_t_13); __pyx_t_13 = 0;
        }
        __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
        __pyx_t_19 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_12); if (unlikely(__pyx_t_19 == -1)) __PYX_ERR(0, 231, __pyx_L6_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;

        /* "fuzzysearch/_generic_search.pyx":229
 *             # note: index == length(sequence)
 *             n_skipped = subseq_len - cand.subseq_index
 *             if cand.n_dels + n_skipped <= max_deletions and \             # <<<<<<<<<<<<<<
 *                cand.l_dist + n_skipped <= max_l_dist:
 *                 matches.append(Match(cand.start, index, cand.l_dist + n_skipped))
 */
      }
    }
  }

  /* "fuzzysearch/_generic_search.pyx":234
 * 
 *     finally:
 *         free(candidates)             # <<<<<<<<<<<<<<
 *         free(new_candidates)
 * 
 */
  /*finally:*/ {
    /*normal exit:*/{
      free(__pyx_v_candidates);

      /* "fuzzysearch/_generic_search.pyx":235
 *     finally:
 *         free(candidates)
 *         free(new_candidates)             # <<<<<<<<<<<<<<
 * 
 *     return matches
 */
      free(__pyx_v_new_candidates);
      goto __pyx_L7;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L6_error:;
      __pyx_t_30 = 0; __pyx_t_31 = 0; __pyx_t_32 = 0; __pyx_t_33 = 0; __pyx_t_34 = 0; __pyx_t_35 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF(__pyx_t_18); __pyx_t_18 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_26); __pyx_t_26 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
      __Pyx_XDECREF(__pyx_t_14); __pyx_t_14 = 0;
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_33, &__pyx_t_34, &__pyx_t_35);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_30, &__pyx_t_31, &__pyx_t_32) < 0)) __Pyx_ErrFetch(&__pyx_t_30, &__pyx_t_31, &__pyx_t_32);
      __Pyx_XGOTREF(__pyx_t_30);
      __Pyx_XGOTREF(__pyx_t_31);
      __Pyx_XGOTREF(__pyx_t_32);
      __Pyx_XGOTREF(__pyx_t_33);
      __Pyx_XGOTREF(__pyx_t_34);
      __Pyx_XGOTREF(__pyx_t_35);
      __pyx_t_17 = __pyx_lineno; __pyx_t_28 = __pyx_clineno; __pyx_t_29 = __pyx_filename;
      {

        /* "fuzzysearch/_generic_search.pyx":234
 * 
 *     finally:
 *         free(candidates)             # <<<<<<<<<<<<<<
 *         free(new_candidates)
 * 
 */
        free(__pyx_v_candidates);

        /* "fuzzysearch/_generic_search.pyx":235
 *     finally:
 *         free(candidates)
 *         free(new_candidates)             # <<<<<<<<<<<<<<
 * 
 *     return matches
 */
        free(__pyx_v_new_candidates);
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_33);
        __Pyx_XGIVEREF(__pyx_t_34);
        __Pyx_XGIVEREF(__pyx_t_35);
        __Pyx_ExceptionReset(__pyx_t_33, __pyx_t_34, __pyx_t_35);
      }
      __Pyx_XGIVEREF(__pyx_t_30);
      __Pyx_XGIVEREF(__pyx_t_31);
      __Pyx_XGIVEREF(__pyx_t_32);
      __Pyx_ErrRestore(__pyx_t_30, __pyx_t_31, __pyx_t_32);
      __pyx_t_30 = 0; __pyx_t_31 = 0; __pyx_t_32 = 0; __pyx_t_33 = 0; __pyx_t_34 = 0; __pyx_t_35 = 0;
      __pyx_lineno = __pyx_t_17; __pyx_clineno = __pyx_t_28; __pyx_filename = __pyx_t_29;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "fuzzysearch/_generic_search.pyx":237
 *         free(new_candidates)
 * 
 *     return matches             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_matches);
  __pyx_r = __pyx_v_matches;
  goto __pyx_L0;

  /* "fuzzysearch/_generic_search.pyx":66
 * # subsequence strings, which means if they contain null bytes the data after
 * # the first null byte will not be copied.
 * cdef _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *         const char* subsequence, size_t subseq_len,
 *         const char* sequence, size_t seq_len,
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_18);
  __Pyx_XDECREF(__pyx_t_26);
  __Pyx_AddTraceback("fuzzysearch._generic_search._c_find_near_matches_generic_linear_programming", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_matches);
  __Pyx_XDECREF(__pyx_v_n_skipped);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "fuzzysearch/_generic_search.pyx":241
 * 
 * 
 * def c_find_near_matches_generic_ngrams(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_11fuzzysearch_15_generic_search_3c_find_near_matches_generic_ngrams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_11fuzzysearch_15_generic_search_2c_find_near_matches_generic_ngrams[] = "search for near-matches of subsequence in sequence\n\n    This searches for near-matches, where the nearly-matching parts of the\n    sequence must meet the following limitations (relative to the subsequence):\n\n    * the maximum allowed number of character substitutions\n    * the maximum allowed number of new characters inserted\n    * and the maximum allowed number of character deletions\n    * the total number of substitutions, insertions and deletions\n    ";
static PyMethodDef __pyx_mdef_11fuzzysearch_15_generic_search_3c_find_near_matches_generic_ngrams = {"c_find_near_matches_generic_ngrams", (PyCFunction)__pyx_pw_11fuzzysearch_15_generic_search_3c_find_near_matches_generic_ngrams, METH_VARARGS|METH_KEYWORDS, __pyx_doc_11fuzzysearch_15_generic_search_2c_find_near_matches_generic_ngrams};
static PyObject *__pyx_pw_11fuzzysearch_15_generic_search_3c_find_near_matches_generic_ngrams(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_subsequence = 0;
  PyObject *__pyx_v_sequence = 0;
  PyObject *__pyx_v_search_params = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("c_find_near_matches_generic_ngrams (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_subsequence,&__pyx_n_s_sequence,&__pyx_n_s_search_params,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_subsequence)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_sequence)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_ngrams", 1, 3, 3, 1); __PYX_ERR(0, 241, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_search_params)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_ngrams", 1, 3, 3, 2); __PYX_ERR(0, 241, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "c_find_near_matches_generic_ngrams") < 0)) __PYX_ERR(0, 241, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v_subsequence = values[0];
    __pyx_v_sequence = values[1];
    __pyx_v_search_params = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("c_find_near_matches_generic_ngrams", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 241, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("fuzzysearch._generic_search.c_find_near_matches_generic_ngrams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_11fuzzysearch_15_generic_search_2c_find_near_matches_generic_ngrams(__pyx_self, __pyx_v_subsequence, __pyx_v_sequence, __pyx_v_search_params);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_11fuzzysearch_15_generic_search_2c_find_near_matches_generic_ngrams(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_subsequence, PyObject *__pyx_v_sequence, PyObject *__pyx_v_search_params) {
  PyObject *__pyx_v_max_substitutions = NULL;
  PyObject *__pyx_v_max_insertions = NULL;
  PyObject *__pyx_v_max_deletions = NULL;
  PyObject *__pyx_v_max_l_dist = NULL;
  size_t __pyx_v__subseq_len;
  CYTHON_UNUSED size_t __pyx_v__subseq_len_minus_one;
  size_t __pyx_v__seq_len;
  unsigned int __pyx_v_c_max_substitutions;
  unsigned int __pyx_v_c_max_insertions;
  unsigned int __pyx_v_c_max_deletions;
  unsigned int __pyx_v_c_max_l_dist;
  char const *__pyx_v_c_sequence;
  char const *__pyx_v_c_subsequence;
  size_t __pyx_v_ngram_len;
  int __pyx_v_small_search_start_index;
  size_t __pyx_v_ngram_start;
  char *__pyx_v_match_ptr;
  int __pyx_v_subseq_sum;
  PyObject *__pyx_v_matches = NULL;
  PyObject *__pyx_v_small_search_length = NULL;
  PyObject *__pyx_v_match = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  Py_ssize_t __pyx_t_10;
  unsigned int __pyx_t_11;
  unsigned int __pyx_t_12;
  char const *__pyx_t_13;
  char const *__pyx_t_14;
  long __pyx_t_15;
  PyObject *(*__pyx_t_16)(PyObject *);
  size_t __pyx_t_17;
  Py_ssize_t __pyx_t_18;
  PyObject *(*__pyx_t_19)(PyObject *);
  PyObject *__pyx_t_20 = NULL;
  int __pyx_t_21;
  __Pyx_RefNannySetupContext("c_find_near_matches_generic_ngrams", 0);

  /* "fuzzysearch/_generic_search.pyx":252
 *     * the total number of substitutions, insertions and deletions
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ALLOWED_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_IsInstance(__pyx_v_sequence, __pyx_t_1); if (unlikely(__pyx_t_2 == -1)) __PYX_ERR(0, 252, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_3) {

    /* "fuzzysearch/_generic_search.pyx":253
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))             # <<<<<<<<<<<<<<
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 */
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_sequence_is_of_invalid_type_s, ((PyObject *)Py_TYPE(__pyx_v_subsequence))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 253, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 253, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":252
 *     * the total number of substitutions, insertions and deletions
 *     """
 *     if not isinstance(sequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 */
  }

  /* "fuzzysearch/_generic_search.pyx":254
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ALLOWED_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = PyObject_IsInstance(__pyx_v_subsequence, __pyx_t_1); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 254, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = ((!(__pyx_t_3 != 0)) != 0);
  if (__pyx_t_2) {

    /* "fuzzysearch/_generic_search.pyx":255
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))             # <<<<<<<<<<<<<<
 * 
 *     if not subsequence:
 */
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_subsequence_is_of_invalid_type_s, ((PyObject *)Py_TYPE(__pyx_v_subsequence))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 255, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 255, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":254
 *     if not isinstance(sequence, ALLOWED_TYPES):
 *         raise TypeError('sequence is of invalid type %s' % type(subsequence))
 *     if not isinstance(subsequence, ALLOWED_TYPES):             # <<<<<<<<<<<<<<
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 */
  }

  /* "fuzzysearch/_generic_search.pyx":257
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 *     if not subsequence:             # <<<<<<<<<<<<<<
 *         raise ValueError('Given subsequence is empty!')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_subsequence); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 257, __pyx_L1_error)
  __pyx_t_3 = ((!__pyx_t_2) != 0);
  if (__pyx_t_3) {

    /* "fuzzysearch/_generic_search.pyx":258
 * 
 *     if not subsequence:
 *         raise ValueError('Given subsequence is empty!')             # <<<<<<<<<<<<<<
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 258, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 258, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":257
 *         raise TypeError('subsequence is of invalid type %s' % type(subsequence))
 * 
 *     if not subsequence:             # <<<<<<<<<<<<<<
 *         raise ValueError('Given subsequence is empty!')
 * 
 */
  }

  /* "fuzzysearch/_generic_search.pyx":260
 *         raise ValueError('Given subsequence is empty!')
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked             # <<<<<<<<<<<<<<
 * 
 *     # optimization: prepare some often used things in advance
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_search_params, __pyx_n_s_unpacked); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 260, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if ((likely(PyTuple_CheckExact(__pyx_t_1))) || (PyList_CheckExact(__pyx_t_1))) {
    PyObject* sequence = __pyx_t_1;
    #if !CYTHON_COMPILING_IN_PYPY
    Py_ssize_t size = Py_SIZE(sequence);
    #else
    Py_ssize_t size = PySequence_Size(sequence);
    #endif
    if (unlikely(size != 4)) {
      if (size > 4) __Pyx_RaiseTooManyValuesError(4);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(0, 260, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (likely(PyTuple_CheckExact(sequence))) {
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyTuple_GET_ITEM(sequence, 2); 
      __pyx_t_7 = PyTuple_GET_ITEM(sequence, 3); 
    } else {
      __pyx_t_4 = PyList_GET_ITEM(sequence, 0); 
      __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
      __pyx_t_6 = PyList_GET_ITEM(sequence, 2); 
      __pyx_t_7 = PyList_GET_ITEM(sequence, 3); 
    }
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_t_6);
    __Pyx_INCREF(__pyx_t_7);
    #else
    {
      Py_ssize_t i;
      PyObject** temps[4] = {&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7};
      for (i=0; i < 4; i++) {
        PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 260, __pyx_L1_error)
        __Pyx_GOTREF(item);
        *(temps[i]) = item;
      }
    }
    #endif
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  } else {
    Py_ssize_t index = -1;
    PyObject** temps[4] = {&__pyx_t_4,&__pyx_t_5,&__pyx_t_6,&__pyx_t_7};
    __pyx_t_8 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 260, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
    for (index=0; index < 4; index++) {
      PyObject* item = __pyx_t_9(__pyx_t_8); if (unlikely(!item)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(item);
      *(temps[index]) = item;
    }
    if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 4) < 0) __PYX_ERR(0, 260, __pyx_L1_error)
    __pyx_t_9 = NULL;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    goto __pyx_L7_unpacking_done;
    __pyx_L6_unpacking_failed:;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_9 = NULL;
    if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
    __PYX_ERR(0, 260, __pyx_L1_error)
    __pyx_L7_unpacking_done:;
  }
  __pyx_v_max_substitutions = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_max_insertions = __pyx_t_5;
  __pyx_t_5 = 0;
  __pyx_v_max_deletions = __pyx_t_6;
  __pyx_t_6 = 0;
  __pyx_v_max_l_dist = __pyx_t_7;
  __pyx_t_7 = 0;

  /* "fuzzysearch/_generic_search.pyx":263
 * 
 *     # optimization: prepare some often used things in advance
 *     cdef size_t _subseq_len = len(subsequence)             # <<<<<<<<<<<<<<
 *     cdef size_t _subseq_len_minus_one = _subseq_len - 1
 *     cdef size_t _seq_len = len(sequence)
 */
  __pyx_t_10 = PyObject_Length(__pyx_v_subsequence); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 263, __pyx_L1_error)
  __pyx_v__subseq_len = __pyx_t_10;

  /* "fuzzysearch/_generic_search.pyx":264
 *     # optimization: prepare some often used things in advance
 *     cdef size_t _subseq_len = len(subsequence)
 *     cdef size_t _subseq_len_minus_one = _subseq_len - 1             # <<<<<<<<<<<<<<
 *     cdef size_t _seq_len = len(sequence)
 * 
 */
  __pyx_v__subseq_len_minus_one = (__pyx_v__subseq_len - 1);

  /* "fuzzysearch/_generic_search.pyx":265
 *     cdef size_t _subseq_len = len(subsequence)
 *     cdef size_t _subseq_len_minus_one = _subseq_len - 1
 *     cdef size_t _seq_len = len(sequence)             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned int c_max_substitutions = max_substitutions if max_substitutions is not None else (1<<29)
 */
  __pyx_t_10 = PyObject_Length(__pyx_v_sequence); if (unlikely(__pyx_t_10 == -1)) __PYX_ERR(0, 265, __pyx_L1_error)
  __pyx_v__seq_len = __pyx_t_10;

  /* "fuzzysearch/_generic_search.pyx":267
 *     cdef size_t _seq_len = len(sequence)
 * 
 *     cdef unsigned int c_max_substitutions = max_substitutions if max_substitutions is not None else (1<<29)             # <<<<<<<<<<<<<<
 *     cdef unsigned int c_max_insertions = max_insertions if max_insertions is not None else (1<<29)
 *     cdef unsigned int c_max_deletions = max_deletions if max_deletions is not None else (1<<29)
 */
  __pyx_t_3 = (__pyx_v_max_substitutions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_12 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_substitutions); if (unlikely((__pyx_t_12 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 267, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_12;
  } else {
    __pyx_t_11 = 0x20000000;
  }
  __pyx_v_c_max_substitutions = __pyx_t_11;

  /* "fuzzysearch/_generic_search.pyx":268
 * 
 *     cdef unsigned int c_max_substitutions = max_substitutions if max_substitutions is not None else (1<<29)
 *     cdef unsigned int c_max_insertions = max_insertions if max_insertions is not None else (1<<29)             # <<<<<<<<<<<<<<
 *     cdef unsigned int c_max_deletions = max_deletions if max_deletions is not None else (1<<29)
 * 
 */
  __pyx_t_3 = (__pyx_v_max_insertions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_12 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_insertions); if (unlikely((__pyx_t_12 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 268, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_12;
  } else {
    __pyx_t_11 = 0x20000000;
  }
  __pyx_v_c_max_insertions = __pyx_t_11;

  /* "fuzzysearch/_generic_search.pyx":269
 *     cdef unsigned int c_max_substitutions = max_substitutions if max_substitutions is not None else (1<<29)
 *     cdef unsigned int c_max_insertions = max_insertions if max_insertions is not None else (1<<29)
 *     cdef unsigned int c_max_deletions = max_deletions if max_deletions is not None else (1<<29)             # <<<<<<<<<<<<<<
 * 
 *     # TODO: write a good comment
 */
  __pyx_t_3 = (__pyx_v_max_deletions != Py_None);
  if ((__pyx_t_3 != 0)) {
    __pyx_t_12 = __Pyx_PyInt_As_unsigned_int(__pyx_v_max_deletions); if (unlikely((__pyx_t_12 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 269, __pyx_L1_error)
    __pyx_t_11 = __pyx_t_12;
  } else {
    __pyx_t_11 = 0x20000000;
  }
  __pyx_v_c_max_deletions = __pyx_t_11;

  /* "fuzzysearch/_generic_search.pyx":274
 *     cdef unsigned int c_max_l_dist = min(
 *         max_l_dist if max_l_dist is not None else (1<<29),
 *         c_max_substitutions + c_max_insertions + c_max_deletions,             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_11 = ((__pyx_v_c_max_substitutions + __pyx_v_c_max_insertions) + __pyx_v_c_max_deletions);

  /* "fuzzysearch/_generic_search.pyx":273
 *     # TODO: write a good comment
 *     cdef unsigned int c_max_l_dist = min(
 *         max_l_dist if max_l_dist is not None else (1<<29),             # <<<<<<<<<<<<<<
 *         c_max_substitutions + c_max_insertions + c_max_deletions,
 *     )
 */
  __pyx_t_3 = (__pyx_v_max_l_dist != Py_None);
  if ((__pyx_t_3 != 0)) {
    __Pyx_INCREF(__pyx_v_max_l_dist);
    __pyx_t_1 = __pyx_v_max_l_dist;
  } else {
    __Pyx_INCREF(__pyx_int_536870912);
    __pyx_t_1 = __pyx_int_536870912;
  }

  /* "fuzzysearch/_generic_search.pyx":274
 *     cdef unsigned int c_max_l_dist = min(
 *         max_l_dist if max_l_dist is not None else (1<<29),
 *         c_max_substitutions + c_max_insertions + c_max_deletions,             # <<<<<<<<<<<<<<
 *     )
 * 
 */
  __pyx_t_6 = __Pyx_PyInt_From_unsigned_int(__pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_5 = PyObject_RichCompare(__pyx_t_6, __pyx_t_1, Py_LT); __Pyx_XGOTREF(__pyx_t_5); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_5); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (__pyx_t_3) {
    __pyx_t_5 = __Pyx_PyInt_From_unsigned_int(__pyx_t_11); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_7 = __pyx_t_5;
    __pyx_t_5 = 0;
  } else {
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_7 = __pyx_t_1;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_11 = __Pyx_PyInt_As_unsigned_int(__pyx_t_7); if (unlikely((__pyx_t_11 == (unsigned int)-1) && PyErr_Occurred())) __PYX_ERR(0, 274, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_v_c_max_l_dist = __pyx_t_11;

  /* "fuzzysearch/_generic_search.pyx":277
 *     )
 * 
 *     cdef const char* c_sequence = sequence             # <<<<<<<<<<<<<<
 *     cdef const char* c_subsequence = subsequence
 * 
 */
  __pyx_t_13 = __Pyx_PyObject_AsString(__pyx_v_sequence); if (unlikely((!__pyx_t_13) && PyErr_Occurred())) __PYX_ERR(0, 277, __pyx_L1_error)
  __pyx_v_c_sequence = __pyx_t_13;

  /* "fuzzysearch/_generic_search.pyx":278
 * 
 *     cdef const char* c_sequence = sequence
 *     cdef const char* c_subsequence = subsequence             # <<<<<<<<<<<<<<
 * 
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)
 */
  __pyx_t_14 = __Pyx_PyObject_AsString(__pyx_v_subsequence); if (unlikely((!__pyx_t_14) && PyErr_Occurred())) __PYX_ERR(0, 278, __pyx_L1_error)
  __pyx_v_c_subsequence = __pyx_t_14;

  /* "fuzzysearch/_generic_search.pyx":280
 *     cdef const char* c_subsequence = subsequence
 * 
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)             # <<<<<<<<<<<<<<
 *     if ngram_len == 0:
 *         raise ValueError('the subsequence length must be greater than max_l_dist')
 */
  __pyx_t_15 = (__pyx_v_c_max_l_dist + 1);
  if (unlikely(__pyx_t_15 == 0)) {
    PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
    __PYX_ERR(0, 280, __pyx_L1_error)
  }
  __pyx_v_ngram_len = (__pyx_v__subseq_len / __pyx_t_15);

  /* "fuzzysearch/_generic_search.pyx":281
 * 
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)
 *     if ngram_len == 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('the subsequence length must be greater than max_l_dist')
 * 
 */
  __pyx_t_3 = ((__pyx_v_ngram_len == 0) != 0);
  if (__pyx_t_3) {

    /* "fuzzysearch/_generic_search.pyx":282
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)
 *     if ngram_len == 0:
 *         raise ValueError('the subsequence length must be greater than max_l_dist')             # <<<<<<<<<<<<<<
 * 
 *     cdef int index, small_search_start_index
 */
    __pyx_t_7 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_Raise(__pyx_t_7, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __PYX_ERR(0, 282, __pyx_L1_error)

    /* "fuzzysearch/_generic_search.pyx":281
 * 
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)
 *     if ngram_len == 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('the subsequence length must be greater than max_l_dist')
 * 
 */
  }

  /* "fuzzysearch/_generic_search.pyx":290
 *     cdef int subseq_sum
 * 
 *     matches = []             # <<<<<<<<<<<<<<
 *     for ngram_start in xrange(0, _subseq_len - ngram_len + 1, ngram_len):
 *         subseq_sum = calc_sum(c_subsequence + ngram_start, ngram_len)
 */
  __pyx_t_7 = PyList_New(0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 290, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_v_matches = ((PyObject*)__pyx_t_7);
  __pyx_t_7 = 0;

  /* "fuzzysearch/_generic_search.pyx":291
 * 
 *     matches = []
 *     for ngram_start in xrange(0, _subseq_len - ngram_len + 1, ngram_len):             # <<<<<<<<<<<<<<
 *         subseq_sum = calc_sum(c_subsequence + ngram_start, ngram_len)
 * 
 */
  __pyx_t_7 = __Pyx_PyInt_FromSize_t(((__pyx_v__subseq_len - __pyx_v_ngram_len) + 1)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_1 = __Pyx_PyInt_FromSize_t(__pyx_v_ngram_len); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_1);
  __pyx_t_7 = 0;
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_xrange, __pyx_t_5, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_5 = __pyx_t_1; __Pyx_INCREF(__pyx_t_5); __pyx_t_10 = 0;
    __pyx_t_16 = NULL;
  } else {
    __pyx_t_10 = -1; __pyx_t_5 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_16 = Py_TYPE(__pyx_t_5)->tp_iternext; if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 291, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_16)) {
      if (likely(PyList_CheckExact(__pyx_t_5))) {
        if (__pyx_t_10 >= PyList_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 291, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_10 >= PyTuple_GET_SIZE(__pyx_t_5)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_5, __pyx_t_10); __Pyx_INCREF(__pyx_t_1); __pyx_t_10++; if (unlikely(0 < 0)) __PYX_ERR(0, 291, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_5, __pyx_t_10); __pyx_t_10++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 291, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_16(__pyx_t_5);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 291, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_17 = __Pyx_PyInt_As_size_t(__pyx_t_1); if (unlikely((__pyx_t_17 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 291, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_ngram_start = __pyx_t_17;

    /* "fuzzysearch/_generic_search.pyx":292
 *     matches = []
 *     for ngram_start in xrange(0, _subseq_len - ngram_len + 1, ngram_len):
 *         subseq_sum = calc_sum(c_subsequence + ngram_start, ngram_len)             # <<<<<<<<<<<<<<
 * 
 *         match_ptr = <char *>simple_memmem_with_needle_sum(
 */
    __pyx_v_subseq_sum = calc_sum((__pyx_v_c_subsequence + __pyx_v_ngram_start), __pyx_v_ngram_len);

    /* "fuzzysearch/_generic_search.pyx":294
 *         subseq_sum = calc_sum(c_subsequence + ngram_start, ngram_len)
 * 
 *         match_ptr = <char *>simple_memmem_with_needle_sum(             # <<<<<<<<<<<<<<
 *             <char *>c_sequence + ngram_start,
 *             _seq_len - ngram_start,
 */
    __pyx_v_match_ptr = ((char *)simple_memmem_with_needle_sum((((char *)__pyx_v_c_sequence) + __pyx_v_ngram_start), (__pyx_v__seq_len - __pyx_v_ngram_start), (((char *)__pyx_v_c_subsequence) + __pyx_v_ngram_start), __pyx_v_ngram_len, __pyx_v_subseq_sum));

    /* "fuzzysearch/_generic_search.pyx":301
 *             subseq_sum)
 * 
 *         while match_ptr != NULL:             # <<<<<<<<<<<<<<
 *             small_search_start_index = (match_ptr - c_sequence) - ngram_start - c_max_l_dist
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)
 */
    while (1) {
      __pyx_t_3 = ((__pyx_v_match_ptr != NULL) != 0);
      if (!__pyx_t_3) break;

      /* "fuzzysearch/_generic_search.pyx":302
 * 
 *         while match_ptr != NULL:
 *             small_search_start_index = (match_ptr - c_sequence) - ngram_start - c_max_l_dist             # <<<<<<<<<<<<<<
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)
 *             if small_search_start_index < 0:
 */
      __pyx_v_small_search_start_index = (((__pyx_v_match_ptr - __pyx_v_c_sequence) - __pyx_v_ngram_start) - __pyx_v_c_max_l_dist);

      /* "fuzzysearch/_generic_search.pyx":303
 *         while match_ptr != NULL:
 *             small_search_start_index = (match_ptr - c_sequence) - ngram_start - c_max_l_dist
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)             # <<<<<<<<<<<<<<
 *             if small_search_start_index < 0:
 *                 small_search_length += small_search_start_index
 */
      __pyx_t_1 = __Pyx_PyInt_FromSize_t((__pyx_v__subseq_len + (2 * __pyx_v_c_max_l_dist))); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 303, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_XDECREF_SET(__pyx_v_small_search_length, __pyx_t_1);
      __pyx_t_1 = 0;

      /* "fuzzysearch/_generic_search.pyx":304
 *             small_search_start_index = (match_ptr - c_sequence) - ngram_start - c_max_l_dist
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)
 *             if small_search_start_index < 0:             # <<<<<<<<<<<<<<
 *                 small_search_length += small_search_start_index
 *                 small_search_start_index = 0
 */
      __pyx_t_3 = ((__pyx_v_small_search_start_index < 0) != 0);
      if (__pyx_t_3) {

        /* "fuzzysearch/_generic_search.pyx":305
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)
 *             if small_search_start_index < 0:
 *                 small_search_length += small_search_start_index             # <<<<<<<<<<<<<<
 *                 small_search_start_index = 0
 *             if small_search_start_index + small_search_length > _seq_len:
 */
        __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_small_search_start_index); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_small_search_length, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 305, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_DECREF_SET(__pyx_v_small_search_length, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "fuzzysearch/_generic_search.pyx":306
 *             if small_search_start_index < 0:
 *                 small_search_length += small_search_start_index
 *                 small_search_start_index = 0             # <<<<<<<<<<<<<<
 *             if small_search_start_index + small_search_length > _seq_len:
 *                 small_search_length = _seq_len - small_search_start_index
 */
        __pyx_v_small_search_start_index = 0;

        /* "fuzzysearch/_generic_search.pyx":304
 *             small_search_start_index = (match_ptr - c_sequence) - ngram_start - c_max_l_dist
 *             small_search_length = _subseq_len + (2 * c_max_l_dist)
 *             if small_search_start_index < 0:             # <<<<<<<<<<<<<<
 *                 small_search_length += small_search_start_index
 *                 small_search_start_index = 0
 */
      }

      /* "fuzzysearch/_generic_search.pyx":307
 *                 small_search_length += small_search_start_index
 *                 small_search_start_index = 0
 *             if small_search_start_index + small_search_length > _seq_len:             # <<<<<<<<<<<<<<
 *                 small_search_length = _seq_len - small_search_start_index
 *             # try to expand left and/or right according to n_ngram
 */
      __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_small_search_start_index); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_1 = PyNumber_Add(__pyx_t_7, __pyx_v_small_search_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __Pyx_PyInt_FromSize_t(__pyx_v__seq_len); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_6 = PyObject_RichCompare(__pyx_t_1, __pyx_t_7, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (__pyx_t_3) {

        /* "fuzzysearch/_generic_search.pyx":308
 *                 small_search_start_index = 0
 *             if small_search_start_index + small_search_length > _seq_len:
 *                 small_search_length = _seq_len - small_search_start_index             # <<<<<<<<<<<<<<
 *             # try to expand left and/or right according to n_ngram
 *             for match in _c_find_near_matches_generic_linear_programming(
 */
        __pyx_t_6 = __Pyx_PyInt_FromSize_t((__pyx_v__seq_len - __pyx_v_small_search_start_index)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 308, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_DECREF_SET(__pyx_v_small_search_length, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "fuzzysearch/_generic_search.pyx":307
 *                 small_search_length += small_search_start_index
 *                 small_search_start_index = 0
 *             if small_search_start_index + small_search_length > _seq_len:             # <<<<<<<<<<<<<<
 *                 small_search_length = _seq_len - small_search_start_index
 *             # try to expand left and/or right according to n_ngram
 */
      }

      /* "fuzzysearch/_generic_search.pyx":313
 *                 c_subsequence, _subseq_len,
 *                 c_sequence + small_search_start_index,
 *                 small_search_length,             # <<<<<<<<<<<<<<
 *                 c_max_substitutions, c_max_insertions, c_max_deletions, c_max_l_dist,
 *             ):
 */
      __pyx_t_17 = __Pyx_PyInt_As_size_t(__pyx_v_small_search_length); if (unlikely((__pyx_t_17 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 313, __pyx_L1_error)

      /* "fuzzysearch/_generic_search.pyx":310
 *                 small_search_length = _seq_len - small_search_start_index
 *             # try to expand left and/or right according to n_ngram
 *             for match in _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *                 c_subsequence, _subseq_len,
 *                 c_sequence + small_search_start_index,
 */
      __pyx_t_6 = __pyx_f_11fuzzysearch_15_generic_search__c_find_near_matches_generic_linear_programming(__pyx_v_c_subsequence, __pyx_v__subseq_len, (__pyx_v_c_sequence + __pyx_v_small_search_start_index), __pyx_t_17, __pyx_v_c_max_substitutions, __pyx_v_c_max_insertions, __pyx_v_c_max_deletions, __pyx_v_c_max_l_dist); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (likely(PyList_CheckExact(__pyx_t_6)) || PyTuple_CheckExact(__pyx_t_6)) {
        __pyx_t_7 = __pyx_t_6; __Pyx_INCREF(__pyx_t_7); __pyx_t_18 = 0;
        __pyx_t_19 = NULL;
      } else {
        __pyx_t_18 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_19 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 310, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      for (;;) {
        if (likely(!__pyx_t_19)) {
          if (likely(PyList_CheckExact(__pyx_t_7))) {
            if (__pyx_t_18 >= PyList_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_18); __Pyx_INCREF(__pyx_t_6); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 310, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          } else {
            if (__pyx_t_18 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_6 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_18); __Pyx_INCREF(__pyx_t_6); __pyx_t_18++; if (unlikely(0 < 0)) __PYX_ERR(0, 310, __pyx_L1_error)
            #else
            __pyx_t_6 = PySequence_ITEM(__pyx_t_7, __pyx_t_18); __pyx_t_18++; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_6);
            #endif
          }
        } else {
          __pyx_t_6 = __pyx_t_19(__pyx_t_7);
          if (unlikely(!__pyx_t_6)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 310, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_match, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "fuzzysearch/_generic_search.pyx":316
 *                 c_max_substitutions, c_max_insertions, c_max_deletions, c_max_l_dist,
 *             ):
 *                 matches.append(match._replace(             # <<<<<<<<<<<<<<
 *                     start=match.start + small_search_start_index,
 *                     end=match.end + small_search_start_index,
 */
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_match, __pyx_n_s_replace); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "fuzzysearch/_generic_search.pyx":317
 *             ):
 *                 matches.append(match._replace(
 *                     start=match.start + small_search_start_index,             # <<<<<<<<<<<<<<
 *                     end=match.end + small_search_start_index,
 *                 ))
 */
        __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_match, __pyx_n_s_start); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_small_search_start_index); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_20 = PyNumber_Add(__pyx_t_4, __pyx_t_8); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_start, __pyx_t_20) < 0) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;

        /* "fuzzysearch/_generic_search.pyx":318
 *                 matches.append(match._replace(
 *                     start=match.start + small_search_start_index,
 *                     end=match.end + small_search_start_index,             # <<<<<<<<<<<<<<
 *                 ))
 *             match_ptr = <char *>simple_memmem_with_needle_sum(
 */
        __pyx_t_20 = __Pyx_PyObject_GetAttrStr(__pyx_v_match, __pyx_n_s_end); if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_20);
        __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_small_search_start_index); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_4 = PyNumber_Add(__pyx_t_20, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 318, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_end, __pyx_t_4) < 0) __PYX_ERR(0, 317, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fuzzysearch/_generic_search.pyx":316
 *                 c_max_substitutions, c_max_insertions, c_max_deletions, c_max_l_dist,
 *             ):
 *                 matches.append(match._replace(             # <<<<<<<<<<<<<<
 *                     start=match.start + small_search_start_index,
 *                     end=match.end + small_search_start_index,
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 316, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        __pyx_t_21 = __Pyx_PyList_Append(__pyx_v_matches, __pyx_t_4); if (unlikely(__pyx_t_21 == -1)) __PYX_ERR(0, 316, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "fuzzysearch/_generic_search.pyx":310
 *                 small_search_length = _seq_len - small_search_start_index
 *             # try to expand left and/or right according to n_ngram
 *             for match in _c_find_near_matches_generic_linear_programming(             # <<<<<<<<<<<<<<
 *                 c_subsequence, _subseq_len,
 *                 c_sequence + small_search_start_index,
 */
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

      /* "fuzzysearch/_generic_search.pyx":320
 *                     end=match.end + small_search_start_index,
 *                 ))
 *             match_ptr = <char *>simple_memmem_with_needle_sum(             # <<<<<<<<<<<<<<
 *                 <char *>match_ptr + 1, _seq_len - (match_ptr - c_sequence) - 1,
 *                 <char *>c_subsequence + ngram_start, ngram_len,
 */
      __pyx_v_match_ptr = ((char *)simple_memmem_with_needle_sum((((char *)__pyx_v_match_ptr) + 1), ((__pyx_v__seq_len - (__pyx_v_match_ptr - __pyx_v_c_sequence)) - 1), (((char *)__pyx_v_c_subsequence) + __pyx_v_ngram_start), __pyx_v_ngram_len, __pyx_v_subseq_sum));
    }

    /* "fuzzysearch/_generic_search.pyx":291
 * 
 *     matches = []
 *     for ngram_start in xrange(0, _subseq_len - ngram_len + 1, ngram_len):             # <<<<<<<<<<<<<<
 *         subseq_sum = calc_sum(c_subsequence + ngram_start, ngram_len)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "fuzzysearch/_generic_search.pyx":325
 *                 subseq_sum);
 * 
 *     return matches             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_matches);
  __pyx_r = __pyx_v_matches;
  goto __pyx_L0;

  /* "fuzzysearch/_generic_search.pyx":241
 * 
 * 
 * def c_find_near_matches_generic_ngrams(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_20);
  __Pyx_AddTraceback("fuzzysearch._generic_search.c_find_near_matches_generic_ngrams", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_max_substitutions);
  __Pyx_XDECREF(__pyx_v_max_insertions);
  __Pyx_XDECREF(__pyx_v_max_deletions);
  __Pyx_XDECREF(__pyx_v_max_l_dist);
  __Pyx_XDECREF(__pyx_v_matches);
  __Pyx_XDECREF(__pyx_v_small_search_length);
  __Pyx_XDECREF(__pyx_v_match);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_generic_search",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_ALLOWED_TYPES, __pyx_k_ALLOWED_TYPES, sizeof(__pyx_k_ALLOWED_TYPES), 0, 0, 1, 1},
  {&__pyx_n_s_Bio_Seq, __pyx_k_Bio_Seq, sizeof(__pyx_k_Bio_Seq), 0, 0, 1, 1},
  {&__pyx_kp_s_C_Users_taleinat_PycharmProjects, __pyx_k_C_Users_taleinat_PycharmProjects, sizeof(__pyx_k_C_Users_taleinat_PycharmProjects), 0, 0, 1, 0},
  {&__pyx_kp_s_Given_subsequence_is_empty, __pyx_k_Given_subsequence_is_empty, sizeof(__pyx_k_Given_subsequence_is_empty), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_n_s_Match, __pyx_k_Match, sizeof(__pyx_k_Match), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_n_s_Seq, __pyx_k_Seq, sizeof(__pyx_k_Seq), 0, 0, 1, 1},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_binary_type, __pyx_k_binary_type, sizeof(__pyx_k_binary_type), 0, 0, 1, 1},
  {&__pyx_n_s_c_find_near_matches_generic_line, __pyx_k_c_find_near_matches_generic_line, sizeof(__pyx_k_c_find_near_matches_generic_line), 0, 0, 1, 1},
  {&__pyx_n_s_c_find_near_matches_generic_ngra, __pyx_k_c_find_near_matches_generic_ngra, sizeof(__pyx_k_c_find_near_matches_generic_ngra), 0, 0, 1, 1},
  {&__pyx_n_s_c_max_deletions, __pyx_k_c_max_deletions, sizeof(__pyx_k_c_max_deletions), 0, 0, 1, 1},
  {&__pyx_n_s_c_max_insertions, __pyx_k_c_max_insertions, sizeof(__pyx_k_c_max_insertions), 0, 0, 1, 1},
  {&__pyx_n_s_c_max_l_dist, __pyx_k_c_max_l_dist, sizeof(__pyx_k_c_max_l_dist), 0, 0, 1, 1},
  {&__pyx_n_s_c_max_substitutions, __pyx_k_c_max_substitutions, sizeof(__pyx_k_c_max_substitutions), 0, 0, 1, 1},
  {&__pyx_n_s_c_sequence, __pyx_k_c_sequence, sizeof(__pyx_k_c_sequence), 0, 0, 1, 1},
  {&__pyx_n_s_c_subsequence, __pyx_k_c_subsequence, sizeof(__pyx_k_c_subsequence), 0, 0, 1, 1},
  {&__pyx_n_s_end, __pyx_k_end, sizeof(__pyx_k_end), 0, 0, 1, 1},
  {&__pyx_n_s_fuzzysearch__generic_search, __pyx_k_fuzzysearch__generic_search, sizeof(__pyx_k_fuzzysearch__generic_search), 0, 0, 1, 1},
  {&__pyx_n_s_fuzzysearch_common, __pyx_k_fuzzysearch_common, sizeof(__pyx_k_fuzzysearch_common), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_index, __pyx_k_index, sizeof(__pyx_k_index), 0, 0, 1, 1},
  {&__pyx_n_s_l_dist, __pyx_k_l_dist, sizeof(__pyx_k_l_dist), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_match, __pyx_k_match, sizeof(__pyx_k_match), 0, 0, 1, 1},
  {&__pyx_n_s_match_ptr, __pyx_k_match_ptr, sizeof(__pyx_k_match_ptr), 0, 0, 1, 1},
  {&__pyx_n_s_matches, __pyx_k_matches, sizeof(__pyx_k_matches), 0, 0, 1, 1},
  {&__pyx_n_s_max_deletions, __pyx_k_max_deletions, sizeof(__pyx_k_max_deletions), 0, 0, 1, 1},
  {&__pyx_n_s_max_insertions, __pyx_k_max_insertions, sizeof(__pyx_k_max_insertions), 0, 0, 1, 1},
  {&__pyx_n_s_max_l_dist, __pyx_k_max_l_dist, sizeof(__pyx_k_max_l_dist), 0, 0, 1, 1},
  {&__pyx_n_s_max_substitutions, __pyx_k_max_substitutions, sizeof(__pyx_k_max_substitutions), 0, 0, 1, 1},
  {&__pyx_n_s_n_dels, __pyx_k_n_dels, sizeof(__pyx_k_n_dels), 0, 0, 1, 1},
  {&__pyx_n_s_n_ins, __pyx_k_n_ins, sizeof(__pyx_k_n_ins), 0, 0, 1, 1},
  {&__pyx_n_s_n_subs, __pyx_k_n_subs, sizeof(__pyx_k_n_subs), 0, 0, 1, 1},
  {&__pyx_n_s_ngram_len, __pyx_k_ngram_len, sizeof(__pyx_k_ngram_len), 0, 0, 1, 1},
  {&__pyx_n_s_ngram_start, __pyx_k_ngram_start, sizeof(__pyx_k_ngram_start), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_replace, __pyx_k_replace, sizeof(__pyx_k_replace), 0, 0, 1, 1},
  {&__pyx_n_s_search_params, __pyx_k_search_params, sizeof(__pyx_k_search_params), 0, 0, 1, 1},
  {&__pyx_n_s_seq_len, __pyx_k_seq_len, sizeof(__pyx_k_seq_len), 0, 0, 1, 1},
  {&__pyx_n_s_sequence, __pyx_k_sequence, sizeof(__pyx_k_sequence), 0, 0, 1, 1},
  {&__pyx_kp_s_sequence_is_of_invalid_type_s, __pyx_k_sequence_is_of_invalid_type_s, sizeof(__pyx_k_sequence_is_of_invalid_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_six, __pyx_k_six, sizeof(__pyx_k_six), 0, 0, 1, 1},
  {&__pyx_n_s_small_search_length, __pyx_k_small_search_length, sizeof(__pyx_k_small_search_length), 0, 0, 1, 1},
  {&__pyx_n_s_small_search_start_index, __pyx_k_small_search_start_index, sizeof(__pyx_k_small_search_start_index), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_subseq_index, __pyx_k_subseq_index, sizeof(__pyx_k_subseq_index), 0, 0, 1, 1},
  {&__pyx_n_s_subseq_len, __pyx_k_subseq_len, sizeof(__pyx_k_subseq_len), 0, 0, 1, 1},
  {&__pyx_n_s_subseq_len_minus_one, __pyx_k_subseq_len_minus_one, sizeof(__pyx_k_subseq_len_minus_one), 0, 0, 1, 1},
  {&__pyx_n_s_subseq_sum, __pyx_k_subseq_sum, sizeof(__pyx_k_subseq_sum), 0, 0, 1, 1},
  {&__pyx_n_s_subsequence, __pyx_k_subsequence, sizeof(__pyx_k_subsequence), 0, 0, 1, 1},
  {&__pyx_kp_s_subsequence_is_of_invalid_type_s, __pyx_k_subsequence_is_of_invalid_type_s, sizeof(__pyx_k_subsequence_is_of_invalid_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_kp_s_the_subsequence_length_must_be_g, __pyx_k_the_subsequence_length_must_be_g, sizeof(__pyx_k_the_subsequence_length_must_be_g), 0, 0, 1, 0},
  {&__pyx_n_s_unpacked, __pyx_k_unpacked, sizeof(__pyx_k_unpacked), 0, 0, 1, 1},
  {&__pyx_n_s_xrange, __pyx_k_xrange, sizeof(__pyx_k_xrange), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(0, 24, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(0, 42, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 47, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(0, 88, __pyx_L1_error)
  #if PY_MAJOR_VERSION >= 3
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_xrange) __PYX_ERR(0, 106, __pyx_L1_error)
  #else
  __pyx_builtin_xrange = __Pyx_GetBuiltinName(__pyx_n_s_xrange); if (!__pyx_builtin_xrange) __PYX_ERR(0, 106, __pyx_L1_error)
  #endif
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "fuzzysearch/_generic_search.pyx":47
 * 
 *     if not subsequence:
 *         raise ValueError('Given subsequence is empty!')             # <<<<<<<<<<<<<<
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked
 */
  __pyx_tuple_ = PyTuple_Pack(1, __pyx_kp_s_Given_subsequence_is_empty); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "fuzzysearch/_generic_search.pyx":258
 * 
 *     if not subsequence:
 *         raise ValueError('Given subsequence is empty!')             # <<<<<<<<<<<<<<
 * 
 *     max_substitutions, max_insertions, max_deletions, max_l_dist = search_params.unpacked
 */
  __pyx_tuple__2 = PyTuple_Pack(1, __pyx_kp_s_Given_subsequence_is_empty); if (unlikely(!__pyx_tuple__2)) __PYX_ERR(0, 258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__2);
  __Pyx_GIVEREF(__pyx_tuple__2);

  /* "fuzzysearch/_generic_search.pyx":282
 *     cdef size_t ngram_len = _subseq_len // (c_max_l_dist + 1)
 *     if ngram_len == 0:
 *         raise ValueError('the subsequence length must be greater than max_l_dist')             # <<<<<<<<<<<<<<
 * 
 *     cdef int index, small_search_start_index
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_the_subsequence_length_must_be_g); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "fuzzysearch/_generic_search.pyx":30
 * 
 * 
 * def c_find_near_matches_generic_linear_programming(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */
  __pyx_tuple__4 = PyTuple_Pack(9, __pyx_n_s_subsequence, __pyx_n_s_sequence, __pyx_n_s_search_params, __pyx_n_s_max_substitutions, __pyx_n_s_max_insertions, __pyx_n_s_max_deletions, __pyx_n_s_max_l_dist, __pyx_n_s_c_subsequence, __pyx_n_s_c_sequence); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);
  __pyx_codeobj__5 = (PyObject*)__Pyx_PyCode_New(3, 0, 9, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__4, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_taleinat_PycharmProjects, __pyx_n_s_c_find_near_matches_generic_line, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__5)) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "fuzzysearch/_generic_search.pyx":241
 * 
 * 
 * def c_find_near_matches_generic_ngrams(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */
  __pyx_tuple__6 = PyTuple_Pack(25, __pyx_n_s_subsequence, __pyx_n_s_sequence, __pyx_n_s_search_params, __pyx_n_s_max_substitutions, __pyx_n_s_max_insertions, __pyx_n_s_max_deletions, __pyx_n_s_max_l_dist, __pyx_n_s_subseq_len, __pyx_n_s_subseq_len_minus_one, __pyx_n_s_seq_len, __pyx_n_s_c_max_substitutions, __pyx_n_s_c_max_insertions, __pyx_n_s_c_max_deletions, __pyx_n_s_c_max_l_dist, __pyx_n_s_c_sequence, __pyx_n_s_c_subsequence, __pyx_n_s_ngram_len, __pyx_n_s_index, __pyx_n_s_small_search_start_index, __pyx_n_s_ngram_start, __pyx_n_s_match_ptr, __pyx_n_s_subseq_sum, __pyx_n_s_matches, __pyx_n_s_small_search_length, __pyx_n_s_match); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);
  __pyx_codeobj__7 = (PyObject*)__Pyx_PyCode_New(3, 0, 25, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__6, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_C_Users_taleinat_PycharmProjects, __pyx_n_s_c_find_near_matches_generic_ngra, 241, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__7)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_536870912 = PyInt_FromLong(536870912L); if (unlikely(!__pyx_int_536870912)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_generic_search(void); /*proto*/
PyMODINIT_FUNC init_generic_search(void)
#else
PyMODINIT_FUNC PyInit__generic_search(void); /*proto*/
PyMODINIT_FUNC PyInit__generic_search(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__generic_search(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_generic_search", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_fuzzysearch___generic_search) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "fuzzysearch._generic_search")) {
      if (unlikely(PyDict_SetItemString(modules, "fuzzysearch._generic_search", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "fuzzysearch/_generic_search.pyx":1
 * import six             # <<<<<<<<<<<<<<
 * from fuzzysearch.common import Match
 * from libc.stdlib cimport malloc, free, realloc
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_six, 0, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_six, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "fuzzysearch/_generic_search.pyx":2
 * import six
 * from fuzzysearch.common import Match             # <<<<<<<<<<<<<<
 * from libc.stdlib cimport malloc, free, realloc
 * 
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_Match);
  __Pyx_GIVEREF(__pyx_n_s_Match);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_Match);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_fuzzysearch_common, __pyx_t_1, -1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_Match); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_Match, __pyx_t_1) < 0) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fuzzysearch/_generic_search.pyx":11
 *                                         int needle_sum)
 * 
 * __all__ = [             # <<<<<<<<<<<<<<
 *     'c_find_near_matches_generic_linear_programming',
 *     'c_find_near_matches_generic_ngrams',
 */
  __pyx_t_2 = PyList_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_INCREF(__pyx_n_s_c_find_near_matches_generic_line);
  __Pyx_GIVEREF(__pyx_n_s_c_find_near_matches_generic_line);
  PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_c_find_near_matches_generic_line);
  __Pyx_INCREF(__pyx_n_s_c_find_near_matches_generic_ngra);
  __Pyx_GIVEREF(__pyx_n_s_c_find_near_matches_generic_ngra);
  PyList_SET_ITEM(__pyx_t_2, 1, __pyx_n_s_c_find_near_matches_generic_ngra);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_2) < 0) __PYX_ERR(0, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fuzzysearch/_generic_search.pyx":21
 * 
 * 
 * ALLOWED_TYPES = (six.binary_type, bytearray)             # <<<<<<<<<<<<<<
 * try:
 *     from Bio.Seq import Seq
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_six); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_binary_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __Pyx_INCREF(((PyObject *)(&PyByteArray_Type)));
  __Pyx_GIVEREF(((PyObject *)(&PyByteArray_Type)));
  PyTuple_SET_ITEM(__pyx_t_2, 1, ((PyObject *)(&PyByteArray_Type)));
  __pyx_t_1 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ALLOWED_TYPES, __pyx_t_2) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fuzzysearch/_generic_search.pyx":22
 * 
 * ALLOWED_TYPES = (six.binary_type, bytearray)
 * try:             # <<<<<<<<<<<<<<
 *     from Bio.Seq import Seq
 * except ImportError:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_5);
    /*try:*/ {

      /* "fuzzysearch/_generic_search.pyx":23
 * ALLOWED_TYPES = (six.binary_type, bytearray)
 * try:
 *     from Bio.Seq import Seq             # <<<<<<<<<<<<<<
 * except ImportError:
 *     pass
 */
      __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_INCREF(__pyx_n_s_Seq);
      __Pyx_GIVEREF(__pyx_n_s_Seq);
      PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_Seq);
      __pyx_t_1 = __Pyx_Import(__pyx_n_s_Bio_Seq, __pyx_t_2, -1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 23, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_Seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 23, __pyx_L2_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_Seq, __pyx_t_2) < 0) __PYX_ERR(0, 23, __pyx_L2_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "fuzzysearch/_generic_search.pyx":22
 * 
 * ALLOWED_TYPES = (six.binary_type, bytearray)
 * try:             # <<<<<<<<<<<<<<
 *     from Bio.Seq import Seq
 * except ImportError:
 */
    }

    /* "fuzzysearch/_generic_search.pyx":27
 *     pass
 * else:
 *     ALLOWED_TYPES += (Seq,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else:*/ {
      __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_ALLOWED_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_Seq); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2);
      PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
      __pyx_t_2 = 0;
      __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PyDict_SetItem(__pyx_d, __pyx_n_s_ALLOWED_TYPES, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L4_except_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    goto __pyx_L9_try_end;
    __pyx_L2_error:;
    __Pyx_PyThreadState_assign
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "fuzzysearch/_generic_search.pyx":24
 * try:
 *     from Bio.Seq import Seq
 * except ImportError:             # <<<<<<<<<<<<<<
 *     pass
 * else:
 */
    __pyx_t_7 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_ImportError);
    if (__pyx_t_7) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L3_exception_handled;
    }
    goto __pyx_L4_except_error;
    __pyx_L4_except_error:;

    /* "fuzzysearch/_generic_search.pyx":22
 * 
 * ALLOWED_TYPES = (six.binary_type, bytearray)
 * try:             # <<<<<<<<<<<<<<
 *     from Bio.Seq import Seq
 * except ImportError:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    goto __pyx_L1_error;
    __pyx_L3_exception_handled:;
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_XGIVEREF(__pyx_t_5);
    __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
    __pyx_L9_try_end:;
  }

  /* "fuzzysearch/_generic_search.pyx":30
 * 
 * 
 * def c_find_near_matches_generic_linear_programming(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11fuzzysearch_15_generic_search_1c_find_near_matches_generic_linear_programming, NULL, __pyx_n_s_fuzzysearch__generic_search); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_find_near_matches_generic_line, __pyx_t_2) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fuzzysearch/_generic_search.pyx":241
 * 
 * 
 * def c_find_near_matches_generic_ngrams(subsequence, sequence, search_params):             # <<<<<<<<<<<<<<
 *     """search for near-matches of subsequence in sequence
 * 
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_11fuzzysearch_15_generic_search_3c_find_near_matches_generic_ngrams, NULL, __pyx_n_s_fuzzysearch__generic_search); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_c_find_near_matches_generic_ngra, __pyx_t_2) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "fuzzysearch/_generic_search.pyx":1
 * import six             # <<<<<<<<<<<<<<
 * from fuzzysearch.common import Match
 * from libc.stdlib cimport malloc, free, realloc
 */
  __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init fuzzysearch._generic_search", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init fuzzysearch._generic_search");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* PyObjectCall */
  #if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyErrFetchRestore */
  #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
  #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* IterFinish */
    static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* PyFunctionFastCall */
    #if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyCFunctionFastCall */
    #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyIntBinop */
    #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* GetException */
    #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* SaveResetException */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* Import */
      static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
      static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* PyErrExceptionMatches */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* CodeObjectCache */
      static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
      #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
      #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
      static CYTHON_INLINE unsigned int __Pyx_PyInt_As_unsigned_int(PyObject *x) {
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 2 * PyLong_SHIFT) {
                            return (unsigned int) (((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 3 * PyLong_SHIFT) {
                            return (unsigned int) (((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) >= 4 * PyLong_SHIFT) {
                            return (unsigned int) (((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned int) ((((((unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) (((unsigned int)-1)*(((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned int) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned int) ((((((((((unsigned int)digits[3]) << PyLong_SHIFT) | (unsigned int)digits[2]) << PyLong_SHIFT) | (unsigned int)digits[1]) << PyLong_SHIFT) | (unsigned int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned int) -1;
        }
    } else {
        unsigned int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned int) -1;
        val = __Pyx_PyInt_As_unsigned_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned int");
    return (unsigned int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned int");
    return (unsigned int) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
    const char neg_one = (char) -1, const_zero = (char) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CheckBinaryVersion */
      static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
      static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
